import pytest

from xdsl.builder import Builder
from xdsl.dialects import arith, builtin, func, memref, scf
from xdsl.dialects.builtin import (
    AffineMapAttr,
    ArrayAttr,
    FloatAttr,
    IndexType,
    IntAttr,
    IntegerType,
    MemRefType,
    NoneAttr,
    StridedLayoutAttr,
    UnrankedMemRefType,
    i32,
    i64,
)
from xdsl.dialects.memref import (
    AllocaOp,
    AllocOp,
    CastOp,
    CopyOp,
    DeallocOp,
    DmaStartOp,
    DmaWaitOp,
    ExtractAlignedPointerAsIndexOp,
    ExtractStridedMetaDataOp,
    LoadOp,
    MemorySpaceCastOp,
    StoreOp,
    SubviewOp,
)
from xdsl.ir import BlockArgument, OpResult
from xdsl.ir.affine import AffineMap
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa
from xdsl.utils.test_value import create_ssa_value


def test_memreftype():
    mem1 = MemRefType(i32, [1])

    assert mem1.get_num_dims() == 1
    assert mem1.get_shape() == (1,)
    assert mem1.element_type is i32

    mem2 = MemRefType(i32, [3, 3, 3])

    assert mem2.get_num_dims() == 3
    assert mem2.get_shape() == (3, 3, 3)
    assert mem2.element_type is i32


def test_memref_load_i32():
    i32_memref_type = MemRefType(i32, [1])
    memref_ssa_value = create_ssa_value(i32_memref_type)
    load = LoadOp.get(memref_ssa_value, [])

    assert load.memref is memref_ssa_value
    assert load.indices == ()
    assert load.res.type is i32


def test_memref_load_i32_with_dimensions():
    i32_memref_type = MemRefType(i32, [2, 3])
    memref_ssa_value = create_ssa_value(i32_memref_type)
    index1 = create_ssa_value(IndexType())
    index2 = create_ssa_value(IndexType())
    load = LoadOp.get(memref_ssa_value, [index1, index2])

    assert load.memref is memref_ssa_value
    assert load.indices[0] is index1
    assert load.indices[1] is index2
    assert load.res.type is i32


def test_memref_store_i32():
    i32_memref_type = MemRefType(i32, [1])
    memref_ssa_value = create_ssa_value(i32_memref_type)
    i32_ssa_value = create_ssa_value(i32)
    store = StoreOp.get(i32_ssa_value, memref_ssa_value, [])

    assert store.memref is memref_ssa_value
    assert store.indices == ()
    assert store.value is i32_ssa_value


def test_memref_store_i32_with_dimensions():
    i32_memref_type = MemRefType(i32, [2, 3])
    memref_ssa_value = create_ssa_value(i32_memref_type)
    i32_ssa_value = create_ssa_value(i32)
    index1 = create_ssa_value(IndexType())
    index2 = create_ssa_value(IndexType())
    store = StoreOp.get(i32_ssa_value, memref_ssa_value, [index1, index2])

    assert store.memref is memref_ssa_value
    assert store.indices[0] is index1
    assert store.indices[1] is index2
    assert store.value is i32_ssa_value


def test_memref_alloc():
    my_i32 = IntegerType(32)
    my_layout = StridedLayoutAttr(strides=(2, 4, 6), offset=8)
    my_memspace = builtin.IntegerAttr(0, i32)
    alloc0 = AllocOp.get(my_i32, 64, [3, 1, 2])
    alloc1 = AllocOp.get(my_i32, 64)
    alloc2 = AllocOp.get(
        my_i32, 64, [3, 1, 2], layout=my_layout, memory_space=my_memspace
    )

    assert alloc0.dynamic_sizes == ()
    assert type(alloc0.results[0]) is OpResult
    assert type(alloc0.results[0].type) is MemRefType
    assert alloc0.results[0].type.get_shape() == (3, 1, 2)
    assert type(alloc1.results[0]) is OpResult
    assert type(alloc1.results[0].type) is MemRefType
    assert alloc1.results[0].type.get_shape() == (1,)
    assert type(alloc2.results[0]) is OpResult
    assert type(alloc2.results[0].type) is MemRefType
    assert alloc2.results[0].type.get_shape() == (3, 1, 2)
    assert alloc2.results[0].type.layout == my_layout
    assert alloc2.results[0].type.memory_space == my_memspace

    dynamic_sizes_1 = [create_ssa_value(builtin.IndexType()) for _ in range(1)]
    dynamic_sizes_3 = [create_ssa_value(builtin.IndexType()) for _ in range(3)]

    alloc4 = AllocOp.get(my_i32, 64, [3, 1, -1], dynamic_sizes=dynamic_sizes_1)
    alloc5 = AllocOp.get(my_i32, 64, [-1, -1, -1], dynamic_sizes=dynamic_sizes_3)
    alloc6 = AllocOp.get(my_i32, 64, [-1, -1, -1], dynamic_sizes=dynamic_sizes_1)

    assert list(alloc4.operands) == dynamic_sizes_1
    assert list(alloc5.operands) == dynamic_sizes_3
    assert list(alloc6.operands) == dynamic_sizes_1

    alloc4.verify()
    alloc5.verify()

    with pytest.raises(
        VerifyException,
        match="op dimension operand count does not equal memref dynamic dimension count",
    ):
        alloc6.verify()


def test_memref_alloca():
    my_i32 = IntegerType(32)
    my_layout = StridedLayoutAttr(strides=(2, 4, 6), offset=8)
    my_memspace = builtin.IntegerAttr(0, i32)
    alloc0 = AllocaOp.get(my_i32, 64, [3, 1, 2])
    alloc1 = AllocaOp.get(my_i32, 64)
    alloc2 = AllocaOp.get(
        my_i32, 64, [3, 1, 2], layout=my_layout, memory_space=my_memspace
    )

    assert type(alloc0.results[0]) is OpResult
    assert type(alloc0.results[0].type) is MemRefType
    assert alloc0.results[0].type.get_shape() == (3, 1, 2)
    assert type(alloc1.results[0]) is OpResult
    assert type(alloc1.results[0].type) is MemRefType
    assert alloc1.results[0].type.get_shape() == (1,)
    assert type(alloc2.results[0]) is OpResult
    assert type(alloc2.results[0].type) is MemRefType
    assert alloc2.results[0].type.get_shape() == (3, 1, 2)
    assert alloc2.results[0].type.layout == my_layout
    assert alloc2.results[0].type.memory_space == my_memspace

    dynamic_sizes_1 = [create_ssa_value(builtin.IndexType()) for _ in range(1)]
    dynamic_sizes_3 = [create_ssa_value(builtin.IndexType()) for _ in range(3)]

    alloc4 = AllocaOp.get(my_i32, 64, [3, 1, -1], dynamic_sizes=dynamic_sizes_1)
    alloc5 = AllocaOp.get(my_i32, 64, [-1, -1, -1], dynamic_sizes=dynamic_sizes_3)
    alloc6 = AllocaOp.get(my_i32, 64, [-1, -1, -1], dynamic_sizes=dynamic_sizes_1)

    assert list(alloc4.operands) == dynamic_sizes_1
    assert list(alloc5.operands) == dynamic_sizes_3
    assert list(alloc6.operands) == dynamic_sizes_1

    alloc4.verify()
    alloc5.verify()

    with pytest.raises(
        VerifyException,
        match="op dimension operand count does not equal memref dynamic dimension count",
    ):
        alloc6.verify()


def test_memref_dealloc():
    my_i32 = IntegerType(32)
    alloc0 = AllocOp.get(my_i32, 64, [3, 1, 2])
    dealloc0 = DeallocOp.get(alloc0)

    assert type(dealloc0.memref) is OpResult


def test_memref_dim():
    idx = arith.ConstantOp.from_int_and_width(1, IndexType())
    alloc0 = AllocOp.get(i32, 64, [3, 1, 2])
    dim_1 = memref.DimOp.from_source_and_index(alloc0, idx)

    assert dim_1.source is alloc0.memref
    assert dim_1.index is idx.result
    assert isinstance(dim_1.result.type, IndexType)


def test_memref_rank():
    alloc0 = AllocOp.get(i32, 64, [3, 1, 2])
    dim_1 = memref.RankOp.from_memref(alloc0)

    assert dim_1.source is alloc0.memref
    assert isinstance(dim_1.rank.type, IndexType)


def test_memref_ExtractAlignedPointerAsIndexOp():
    ref = AllocOp.get(i32, 64, [64, 64, 64])
    ptr = ExtractAlignedPointerAsIndexOp.get(ref)

    assert ptr.aligned_pointer.type == IndexType()
    assert ptr.source == ref.memref


def test_memref_matmul_verify():
    memref_f64_rank2 = memref.MemRefType(builtin.f64, [-1, -1])

    @builtin.ModuleOp
    @Builder.implicit_region
    def module():
        @Builder.implicit_region((memref_f64_rank2, memref_f64_rank2))
        def matmul(args: tuple[BlockArgument, ...]) -> None:
            a, b = args

            lit0 = arith.ConstantOp.from_int_and_width(0, builtin.IndexType())
            lit1 = arith.ConstantOp.from_int_and_width(1, builtin.IndexType())
            dim_a0 = memref.DimOp.from_source_and_index(a, lit0)
            dim_a1 = memref.DimOp.from_source_and_index(a, lit1)
            dim_b0 = memref.DimOp.from_source_and_index(b, lit0)
            dim_b1 = memref.DimOp.from_source_and_index(b, lit1)
            out = memref.AllocaOp.get(builtin.f64, 0, [-1, -1], [dim_a0, dim_b1])
            # TODO: assert dim_a0 == dim_b1
            lit0_f = arith.ConstantOp(FloatAttr(0.0, builtin.f64))

            @Builder.implicit_region((builtin.IndexType(),))
            def outer_loop(args: tuple[BlockArgument, ...]):
                (i,) = args

                # outer loop start, loop_var = i

                @Builder.implicit_region((builtin.IndexType(),))
                def mid_loop(args: tuple[BlockArgument, ...]):
                    (j,) = args
                    # mid loop start, loop_var = j
                    memref.StoreOp.get(lit0_f, out, [i, j])

                    @Builder.implicit_region((builtin.IndexType(),))
                    def inner_loop(args: tuple[BlockArgument, ...]):
                        (k,) = args
                        # inner loop, loop_var = k
                        elem_a_i_k = memref.LoadOp.get(a, [i, k])
                        elem_b_k_j = memref.LoadOp.get(b, [k, j])
                        mul = arith.MulfOp(elem_a_i_k, elem_b_k_j)
                        out_i_j = memref.LoadOp.get(out, [i, j])
                        new_out_val = arith.AddfOp(out_i_j, mul)
                        memref.StoreOp.get(new_out_val, out, [i, j])
                        scf.YieldOp()

                    scf.ForOp(lit0, dim_a1, lit1, [], inner_loop)
                    scf.YieldOp()

                scf.ForOp(lit0, dim_b0, lit1, [], mid_loop)
                scf.YieldOp()

            scf.ForOp(lit0, dim_a0, lit1, [], outer_loop)

            func.ReturnOp(out)

        func.FuncOp(
            "matmul",
            ((memref_f64_rank2, memref_f64_rank2), (memref_f64_rank2,)),
            matmul,
        )

    # check that it verifies correctly
    module.verify()


def test_memref_extract_strided_metadata():
    # input type
    el_type = i32
    alignment = 8
    shape = [10, 10, 10]
    strides = [1, 2, 2]
    offset = None  # Dynamic offset
    layout = StridedLayoutAttr(strides, offset)
    memory_space = IntAttr(1)
    alloc = AllocOp.get(
        el_type, alignment, shape, layout=layout, memory_space=memory_space
    )
    assert isa(alloc.memref.type, MemRefType)

    extract_op = ExtractStridedMetaDataOp(alloc)
    # output type is expected to have no layout information, and no shape
    expected_type = MemRefType(el_type, [], NoneAttr(), memory_space=memory_space)
    assert extract_op.results[0].type == expected_type


def test_memref_subview_constant_parameters():
    alloc = AllocOp.get(i32, 8, [10, 10, 10])
    assert isa(alloc.memref.type, MemRefType)

    subview = SubviewOp.from_static_parameters(
        alloc, alloc.memref.type, [2, 2, 2], [2, 2, 2], [3, 3, 3]
    )

    assert isinstance(subview, SubviewOp)
    assert isinstance(subview.result.type, MemRefType)
    assert isinstance(subview.result.type.layout, StridedLayoutAttr)
    assert isa(subview.result.type.layout.strides, ArrayAttr[IntAttr])
    out_strides = [i.data for i in subview.result.type.layout.strides.data]
    assert out_strides == [300, 30, 3]
    assert isinstance(subview.result.type.layout.offset, IntAttr)
    assert subview.result.type.layout.offset.data == 222


def test_memref_cast():
    i32_memref_type = MemRefType(i32, [10, 2])
    memref_ssa_value = create_ssa_value(i32_memref_type)

    res_type = UnrankedMemRefType.from_type(i32)

    cast = CastOp.get(memref_ssa_value, res_type)

    assert cast.source is memref_ssa_value
    assert cast.dest.type is res_type


def test_memref_memory_space_cast():
    i32_memref_type = MemRefType(i32, [10, 2], memory_space=builtin.IntegerAttr(1, i32))
    memref_ssa_value = create_ssa_value(i32_memref_type)

    res_type = MemRefType(i32, [10, 2], memory_space=builtin.IntegerAttr(2, i32))
    res_type_wrong_type = MemRefType(
        i64, [10, 2], memory_space=builtin.IntegerAttr(2, i32)
    )
    res_type_wrong_shape = MemRefType(
        i32, [10, 4], memory_space=builtin.IntegerAttr(2, i32)
    )

    memory_space_cast = MemorySpaceCastOp(memref_ssa_value, res_type)

    assert memory_space_cast.source is memref_ssa_value
    assert memory_space_cast.dest.type is res_type

    with pytest.raises(
        VerifyException,
        match="Expected source and destination to have the same element type.",
    ):
        MemorySpaceCastOp(memref_ssa_value, res_type_wrong_type).verify()

    with pytest.raises(
        VerifyException, match="Expected source and destination to have the same shape."
    ):
        MemorySpaceCastOp(memref_ssa_value, res_type_wrong_shape).verify()

    # Test helper function
    dest_memory_space = builtin.IntegerAttr(2, i32)
    memory_space_cast = MemorySpaceCastOp.from_type_and_target_space(
        memref_ssa_value, i32_memref_type, dest_memory_space
    )

    assert memory_space_cast.source is memref_ssa_value
    assert isinstance(memory_space_cast.dest.type, MemRefType)
    assert memory_space_cast.dest.type.memory_space is dest_memory_space


def test_dma_start():
    src_type = MemRefType(i64, [4, 512], memory_space=IntAttr(1))
    dest_type = MemRefType(i64, [4 * 512], memory_space=IntAttr(2))

    tag_type = MemRefType(i32, [4])

    src = create_ssa_value(src_type)
    dest = create_ssa_value(dest_type)
    tag = create_ssa_value(tag_type)

    index = create_ssa_value(IndexType())
    num_elements = create_ssa_value(IndexType())

    dma_start = DmaStartOp.get(
        src, [index, index], dest, [index], num_elements, tag, [index]
    )

    dma_start.verify()

    # check that src index count is verified
    with pytest.raises(VerifyException, match="Expected 2 source indices"):
        DmaStartOp.get(
            src, [index, index, index], dest, [index], num_elements, tag, [index]
        ).verify()

    # check that dest index count is verified
    with pytest.raises(VerifyException, match="Expected 1 dest indices"):
        DmaStartOp.get(
            src, [index, index], dest, [], num_elements, tag, [index]
        ).verify()

    # check that tag index count is verified
    with pytest.raises(VerifyException, match="Expected 1 tag indices"):
        DmaStartOp.get(
            src, [index, index], dest, [index], num_elements, tag, [index, index]
        ).verify()

    # check that tag index count is verified
    with pytest.raises(VerifyException, match="different memory spaces"):
        DmaStartOp.get(
            src, [index, index], src, [index, index], num_elements, tag, [index]
        ).verify()

    # check that tag element type is verified
    with pytest.raises(VerifyException, match="Expected tag to be a memref of i32"):
        new_tag = create_ssa_value(src_type)

        DmaStartOp.get(
            src,
            [index, index],
            dest,
            [index],
            num_elements,
            new_tag,
            [index, index],
        ).verify()


def test_memref_dma_wait():
    tag_type = MemRefType(i32, [4])
    tag = create_ssa_value(tag_type)
    index = create_ssa_value(IndexType())
    num_elements = create_ssa_value(IndexType())

    dma_wait = DmaWaitOp.get(tag, [index], num_elements)

    dma_wait.verify()

    # check that tag index count is verified
    with pytest.raises(
        VerifyException, match="Expected 1 tag indices because of shape of tag memref"
    ):
        DmaWaitOp.get(tag, [index, index], num_elements).verify()

    # check that tag element type is verified
    with pytest.raises(VerifyException, match="Expected tag to be a memref of i32"):
        wrong_tag_type = MemRefType(i64, [4])
        wrong_tag = create_ssa_value(wrong_tag_type)

        DmaWaitOp.get(wrong_tag, [index], num_elements).verify()


def test_memref_copy():
    i32type4 = MemRefType(i32, [4])
    i32type3 = MemRefType(i32, [3])
    i64type4 = MemRefType(i64, [4])
    source = create_ssa_value(i32type4)
    destination = create_ssa_value(i32type4)

    copy = CopyOp(source, destination)

    copy.verify()
    assert isinstance(copy, CopyOp)
    assert copy.source.type == i32type4
    assert copy.destination.type == i32type4

    destination = create_ssa_value(i32type3)

    copy = CopyOp(source, destination)

    with pytest.raises(
        VerifyException, match="Expected source and destination to have the same shape."
    ):
        copy.verify()

    destination = create_ssa_value(i64type4)

    copy = CopyOp(source, destination)

    with pytest.raises(
        VerifyException,
        match="Expected source and destination to have the same element type.",
    ):
        copy.verify()


def test_get_strides():
    strided = StridedLayoutAttr((24, 4, 1), 5)
    affine = AffineMapAttr(AffineMap.identity(3))

    assert tuple(strided.get_strides()) == (24, 4, 1)
    assert affine.get_strides() is None

    t_none = MemRefType(i32, (2, 3, 4))
    assert (strides := t_none.get_strides())
    assert tuple(strides) == (12, 4, 1)

    t_id = MemRefType(i32, (2, 3, 4), strided)
    assert (strides := t_id.get_strides())
    assert tuple(strides) == (24, 4, 1)



from xdsl.builder import ImplicitBuilder
from xdsl.dialects import func, ml_program
from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    ModuleOp,
    StringAttr,
    SymbolRefAttr,
    TensorType,
    i32,
    i64,
)
from xdsl.interpreter import Interpreter
from xdsl.interpreters.ml_program import MLProgramFunctions
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils.ptr import TypedPtr


def test_ml_program_global_load_constant():
    tensor_type = TensorType(i32, [4])
    module = ModuleOp([])
    with ImplicitBuilder(module.body):
        ml_program.GlobalOp(
            StringAttr("my_global"),
            tensor_type,
            None,
            DenseIntOrFPElementsAttr.from_list(tensor_type, [4]),
            StringAttr("private"),
        )
        with ImplicitBuilder(func.FuncOp("main", ((), ())).body):
            fetch = ml_program.GlobalLoadConstantOp(
                SymbolRefAttr("my_global"), tensor_type
            )

    interpreter = Interpreter(module)
    interpreter.register_implementations(MLProgramFunctions())

    (result,) = interpreter.run_op(fetch, ())
    assert result == ShapedArray(TypedPtr.new_int32([4] * 4), [4])


def test_ml_program_global_load_constant_ex2():
    tensor_type = TensorType(i64, [2])
    module = ModuleOp([])
    with ImplicitBuilder(module.body):
        ml_program.GlobalOp(
            StringAttr("my_global"),
            tensor_type,
            None,
            DenseIntOrFPElementsAttr.from_list(tensor_type, [1, 320]),
            StringAttr("private"),
        )
        with ImplicitBuilder(func.FuncOp("main", ((), ())).body):
            fetch = ml_program.GlobalLoadConstantOp(
                SymbolRefAttr("my_global"), tensor_type
            )

    interpreter = Interpreter(module)
    interpreter.register_implementations(MLProgramFunctions())

    (result,) = interpreter.run_op(fetch, ())
    assert result == ShapedArray(TypedPtr.new_int64([1, 320]), [2])



"""
A collection of helpers for reinterpreting bits.
Used in lowering and interpreting low-level dialects.
"""

import ctypes
import struct


def convert_f16_to_u16(value: float) -> int:
    """
    Convert an IEEE 754 float to a raw unsigned integer representation.
    """
    # using struct library as ctypes does not support half-precision floats
    return struct.unpack("<H", struct.pack("<e", value))[0]


def convert_u16_to_f16(value: int) -> int:
    """
    Convert an IEEE 754 float to a raw unsigned integer representation.
    """
    # using struct library as ctypes does not support half-precision floats
    return struct.unpack("<e", struct.pack("<H", value))[0]


def convert_f32_to_u32(value: float) -> int:
    """
    Convert an IEEE 754 float to a raw unsigned integer representation.
    """
    raw_float = ctypes.c_float(value)
    raw_int = ctypes.c_uint32.from_buffer(raw_float).value
    return raw_int


def convert_u32_to_f32(value: int) -> float:
    """
    Convert a raw 32-bit unsigned integer to IEEE 754 float representation.
    """
    raw_int = ctypes.c_uint32(value)
    raw_float = ctypes.c_float.from_buffer(raw_int).value
    return raw_float


def convert_f64_to_u64(value: float) -> int:
    """
    Convert an IEEE 754 float to a raw unsigned integer representation.
    """
    raw_float = ctypes.c_double(value)
    raw_int = ctypes.c_uint64.from_buffer(raw_float).value
    return raw_int


def convert_u64_to_f64(value: int) -> float:
    """
    Convert a raw 32-bit unsigned integer to IEEE 754 float representation.
    """
    raw_int = ctypes.c_uint64(value)
    raw_float = ctypes.c_double.from_buffer(raw_int).value
    return raw_float


def is_power_of_two(value: int) -> bool:
    """
    Return True if an integer is a power of two.
    Powers of two have only one bit set to one
    """
    return (value > 0) and (value.bit_count() == 1)



from __future__ import annotations

import itertools
from collections.abc import Iterator, Sequence
from dataclasses import KW_ONLY, dataclass, field
from typing import Generic, Literal

from typing_extensions import Self, TypeVar

from xdsl.dialects.builtin import (
    Float32Type,
    Float64Type,
    PackableType,
    i32,
    i64,
)

_T = TypeVar("_T")


@dataclass
class RawPtr:
    """
    Data structure to help simulate pointers into memory.
    """

    memory: bytearray
    offset: int = field(default=0)

    @property
    def memoryview(self) -> memoryview:
        return memoryview(self.memory)[self.offset :]

    def copy(self) -> RawPtr:
        return RawPtr(bytearray(self.memory), self.offset)

    @staticmethod
    def zeros(count: int) -> RawPtr:
        """
        Returns a new Ptr of size `count` with offset 0.
        """
        return RawPtr(bytearray(count))

    def __add__(self, offset: int) -> RawPtr:
        """
        Aliases the data, so storing into the offset stores for all other references
        to the list.
        """
        return RawPtr(self.memory, self.offset + offset)

    @property
    def int32(self) -> TypedPtr[int]:
        return TypedPtr(self, xtype=int32)

    @property
    def int64(self) -> TypedPtr[int]:
        return TypedPtr(self, xtype=int64)

    def index(self, index_bitwidth: int) -> TypedPtr[int]:
        if index_bitwidth != 32 and index_bitwidth != 64:
            raise ValueError(
                f"Invalid index bitwidth {index_bitwidth} monly 32 or 64 allowed"
            )
        return TypedPtr(self, xtype=index(index_bitwidth))

    @property
    def float32(self) -> TypedPtr[float]:
        return TypedPtr(self, xtype=float32)

    @property
    def float64(self) -> TypedPtr[float]:
        return TypedPtr(self, xtype=Float64Type())


int32 = i32
int64 = i64
float32 = Float32Type()
float64 = Float64Type()


def index(bitwidth: Literal[32, 64]) -> PackableType[int]:
    return int32 if bitwidth == 32 else int64


@dataclass
class TypedPtr(Generic[_T]):
    """
    A typed pointer into memory, similar to numpy's ndarray, but without the shape.
    """

    raw: RawPtr
    _: KW_ONLY
    xtype: PackableType[_T]

    @property
    def size(self) -> int:
        return self.xtype.compile_time_size

    def copy(self) -> Self:
        return type(self)(self.raw.copy(), xtype=self.xtype)

    def get_iter(self) -> Iterator[_T]:
        # The memoryview needs to be a multiple of the size of the packed format
        format_size = self.size
        mem_view = self.raw.memoryview
        remainder = len(mem_view) % format_size
        if remainder:
            mem_view = mem_view[:-remainder]
        return self.xtype.iter_unpack(mem_view)

    def get_list(self, count: int) -> list[_T]:
        return list(itertools.islice(self.get_iter(), count))

    def __getitem__(self, index: int) -> _T:
        raw_at_offset = self.raw + index * self.size
        return self.xtype.unpack(raw_at_offset.memoryview[: self.size], 1)[0]

    def __setitem__(self, index: int, value: _T):
        raw_at_offset = self.raw + index * self.size
        self.xtype.pack_into(raw_at_offset.memory, raw_at_offset.offset, value)

    @staticmethod
    def zeros(count: int, *, xtype: PackableType[_T]) -> TypedPtr[_T]:
        size = xtype.compile_time_size
        return TypedPtr(RawPtr.zeros(size * count), xtype=xtype)

    @staticmethod
    def new(els: Sequence[_T], *, xtype: PackableType[_T]) -> TypedPtr[_T]:
        """
        Returns a new TypedPtr with the specified els packed into memory.
        """
        el_size = xtype.compile_time_size
        res = RawPtr.zeros(len(els) * el_size)
        for i, el in enumerate(els):
            xtype.pack_into(res.memory, i * el_size, el)
        return TypedPtr(res, xtype=xtype)

    @staticmethod
    def new_float32(els: Sequence[float]) -> TypedPtr[float]:
        return TypedPtr[float].new(els, xtype=Float32Type())

    @staticmethod
    def new_float64(els: Sequence[float]) -> TypedPtr[float]:
        return TypedPtr[float].new(els, xtype=Float64Type())

    @staticmethod
    def new_int32(els: Sequence[int]) -> TypedPtr[int]:
        return TypedPtr[int].new(els, xtype=i32)

    @staticmethod
    def new_int64(els: Sequence[int]) -> TypedPtr[int]:
        return TypedPtr[int].new(els, xtype=i64)

    @staticmethod
    def new_index(els: Sequence[int], index_bitwidth: int) -> TypedPtr[int]:
        if index_bitwidth != 32 and index_bitwidth != 64:
            raise ValueError(
                f"Invalid index bitwidth {index_bitwidth} monly 32 or 64 allowed"
            )
        return TypedPtr[int].new(els, xtype=index(index_bitwidth))



# RUN: python %s | filecheck %s

from collections.abc import Callable

from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.program import FrontendProgram
from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException

p = FrontendProgram()
try:
    with CodeContext(p):
        # CHECK: Else clause in for loops is not supported.
        def test_not_supported_loop_I():
            for _i in range(10):
                pass
            else:
                pass
            return

    p.compile(desymref=False)
    exit(1)
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Only range-based loops are supported.
        def test_not_supported_loop_II():
            for _i, _j in enumerate(range(10, 0, -1)):
                pass
            return

    p.compile(desymref=False)
    exit(1)
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Range-based loop expected between 1 and 3 arguments, but got 4.
        def test_not_supported_loop_III():
            for _i in range(0, 1, 2, 3):  # pyright: ignore[reportCallIssue]
                pass
            return

    p.compile(desymref=False)
    exit(1)
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Range-based loop must take a single target variable, e.g. `for i in range(10)`.
        def test_not_supported_loop_IV():
            for _i, _j in range(100):  # pyright: ignore[reportGeneralTypeIssues, reportUnknownVariableType]
                pass
            return

    p.compile(desymref=False)
    exit(1)
except CodeGenerationException as e:
    print(e.msg)


try:
    with CodeContext(p):
        # CHECK: Unsupported function argument type: 'Callable[..., None]'
        def test_complex_arg_annotation(x: Callable[..., None]) -> None:
            return

    p.compile(desymref=False)
    exit(1)
except CodeGenerationException as e:
    print(e.msg)


try:
    with CodeContext(p):
        # CHECK: Unsupported function return type: 'int | None'
        def test_complex_return_annotation() -> int | None:
            return

    p.compile(desymref=False)
    exit(1)
except CodeGenerationException as e:
    print(e.msg)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import IO, Any

from xdsl.context import Context
from xdsl.dialects import pdl
from xdsl.dialects.builtin import IntegerAttr, IntegerType, ModuleOp
from xdsl.interpreter import Interpreter, InterpreterFunctions, impl, register_impls
from xdsl.ir import Attribute, Operation, OpResult, SSAValue, TypeAttribute
from xdsl.irdl import IRDLOperation
from xdsl.pattern_rewriter import PatternRewriter, RewritePattern
from xdsl.utils.exceptions import InterpretationError
from xdsl.utils.hints import isa


@dataclass
class PDLMatcher:
    """
    Tracks the xDSL values corresponding to PDL SSA values during
    interpretation. A new instance is created per operation being checked
    against.
    """

    matching_context: dict[SSAValue, Operation | Attribute | SSAValue] = field(
        default_factory=dict[SSAValue, Operation | Attribute | SSAValue]
    )
    """
    For each SSAValue that is an OpResult of an operation in the PDL dialect,
    the corresponding xDSL object.
    """

    native_constraints: dict[str, Callable[..., bool]] = field(
        default_factory=lambda: {}
    )
    """
    The functions that can be used in `pdl.apply_native_constraint`. Note that we do
    not verify that the functions are used with the correct types.
    """

    def get_constant_or_matched_value(
        self, ssa_val: SSAValue
    ) -> Operation | Attribute | SSAValue:
        """
        Get the value that is already matched, or that is defined by a constant such as
        the result of a constant `pdl.attribute`, or of a `pdl.type`, or of a matched
        operation.
        """
        if ssa_val in self.matching_context:
            return self.matching_context[ssa_val]
        if isinstance(ssa_val.owner, pdl.AttributeOp):
            if ssa_val.owner.value is None:
                raise InterpretationError("expected constant `pdl.attribute`")
            return ssa_val.owner.value
        if isinstance(ssa_val.owner, pdl.TypeOp):
            if ssa_val.owner.constantType is None:
                raise InterpretationError("expected constant `pdl.type`")
            return ssa_val.owner.constantType
        raise InterpretationError("expected constant or matched value")

    def match_operand(
        self, ssa_val: SSAValue, pdl_op: pdl.OperandOp, xdsl_val: SSAValue
    ):
        if ssa_val in self.matching_context:
            return self.matching_context[ssa_val] == xdsl_val

        if pdl_op.value_type is not None:
            assert isinstance(pdl_op.value_type, OpResult)
            assert isinstance(pdl_op.value_type.op, pdl.TypeOp)

            if not self.match_type(
                pdl_op.value_type, pdl_op.value_type.op, xdsl_val.type
            ):
                return False

        self.matching_context[ssa_val] = xdsl_val

        return True

    def match_result(
        self, ssa_val: SSAValue, pdl_op: pdl.ResultOp, xdsl_operand: SSAValue
    ):
        if ssa_val in self.matching_context:
            return self.matching_context[ssa_val] == xdsl_operand

        root_pdl_op_value = pdl_op.parent_
        assert isinstance(root_pdl_op_value, OpResult)
        assert isinstance(root_pdl_op_value.op, pdl.OperationOp)

        if not isinstance(xdsl_operand, OpResult):
            return False

        xdsl_op = xdsl_operand.op

        if not self.match_operation(root_pdl_op_value, root_pdl_op_value.op, xdsl_op):
            return False

        original_op = root_pdl_op_value.op

        index = pdl_op.index.value.data

        if len(original_op.results) <= index:
            return False

        self.matching_context[ssa_val] = xdsl_op.results[index]

        return True

    def match_type(self, ssa_val: SSAValue, pdl_op: pdl.TypeOp, xdsl_attr: Attribute):
        if ssa_val in self.matching_context:
            return self.matching_context[ssa_val] == xdsl_attr

        if pdl_op.constantType is None or pdl_op.constantType == xdsl_attr:
            self.matching_context[ssa_val] = xdsl_attr
            return True
        else:
            return False

    def match_attribute(
        self,
        ssa_val: SSAValue,
        pdl_op: pdl.AttributeOp,
        attr_name: str,
        xdsl_attr: Attribute,
    ):
        if ssa_val in self.matching_context:
            return self.matching_context[ssa_val] == xdsl_attr

        if pdl_op.value is not None:
            if pdl_op.value != xdsl_attr:
                return False

        if pdl_op.value_type is not None:
            assert isinstance(pdl_op.value_type, OpResult)
            assert isinstance(pdl_op.value_type.op, pdl.TypeOp)

            assert isa(xdsl_attr, IntegerAttr[IntegerType]), (
                "Only handle integer types for now"
            )

            if not self.match_type(
                pdl_op.value_type, pdl_op.value_type.op, xdsl_attr.type
            ):
                return False

        self.matching_context[ssa_val] = xdsl_attr

        return True

    def match_operation(
        self, ssa_val: SSAValue, pdl_op: pdl.OperationOp, xdsl_op: Operation
    ) -> bool:
        if ssa_val in self.matching_context:
            return self.matching_context[ssa_val] == xdsl_op

        if pdl_op.opName is not None:
            if xdsl_op.name != pdl_op.opName.data:
                return False

        attribute_value_names = [avn.data for avn in pdl_op.attributeValueNames.data]

        for avn, av in zip(attribute_value_names, pdl_op.attribute_values):
            assert isinstance(av, OpResult)
            assert isinstance(av.op, pdl.AttributeOp)
            if (attr := xdsl_op.get_attr_or_prop(avn)) is None:
                return False

            if not self.match_attribute(av, av.op, avn, attr):
                return False

        pdl_operands = pdl_op.operand_values
        xdsl_operands = xdsl_op.operands

        if len(pdl_operands) != len(xdsl_operands):
            return False

        for pdl_operand, xdsl_operand in zip(pdl_operands, xdsl_operands):
            assert isinstance(pdl_operand, OpResult)
            assert isinstance(pdl_operand.op, pdl.OperandOp | pdl.ResultOp)
            match pdl_operand.op:
                case pdl.OperandOp():
                    if not self.match_operand(
                        pdl_operand, pdl_operand.op, xdsl_operand
                    ):
                        return False
                case pdl.ResultOp():
                    if not self.match_result(pdl_operand, pdl_operand.op, xdsl_operand):
                        return False

        pdl_results = pdl_op.type_values
        xdsl_results = xdsl_op.results

        if len(pdl_results) != len(xdsl_results):
            return False

        for pdl_result, xdsl_result in zip(pdl_results, xdsl_results):
            assert isinstance(pdl_result, OpResult)
            assert isinstance(pdl_result.op, pdl.TypeOp)
            if not self.match_type(pdl_result, pdl_result.op, xdsl_result.type):
                return False

        self.matching_context[ssa_val] = xdsl_op

        return True

    def check_native_constraints(self, pdl_op: pdl.ApplyNativeConstraintOp) -> bool:
        args = [
            self.get_constant_or_matched_value(operand) for operand in pdl_op.operands
        ]
        name = pdl_op.constraint_name.data
        if name not in self.native_constraints:
            raise InterpretationError(f"{name} PDL native constraint is not registered")
        return self.native_constraints[name](*args)


@dataclass
class PDLRewritePattern(RewritePattern):
    functions: PDLRewriteFunctions
    pdl_rewrite_op: pdl.RewriteOp
    interpreter: Interpreter
    native_constraints: dict[str, Callable[..., bool]]

    def __init__(
        self,
        pdl_rewrite_op: pdl.RewriteOp,
        ctx: Context,
        file: IO[str] | None = None,
        native_constraints: dict[str, Callable[..., bool]] | None = None,
    ):
        pdl_pattern = pdl_rewrite_op.parent_op()
        assert isinstance(pdl_pattern, pdl.PatternOp)
        pdl_module = pdl_pattern.parent_op()
        assert isinstance(pdl_module, ModuleOp)
        self.functions = PDLRewriteFunctions(ctx)
        self.interpreter = Interpreter(pdl_module, file=file)
        self.interpreter.register_implementations(self.functions)
        self.pdl_rewrite_op = pdl_rewrite_op
        if native_constraints is None:
            native_constraints = {}
        self.native_constraints = native_constraints

    def match_and_rewrite(self, xdsl_op: Operation, rewriter: PatternRewriter) -> None:
        pdl_op_val = self.pdl_rewrite_op.root
        assert pdl_op_val is not None, "TODO: handle None root op in pdl.RewriteOp"
        assert self.pdl_rewrite_op.body is not None, (
            "TODO: handle None body op in pdl.RewriteOp"
        )

        assert isinstance(pdl_op_val, OpResult)
        pdl_op = pdl_op_val.op

        assert isinstance(pdl_op, pdl.OperationOp)
        matcher = PDLMatcher(native_constraints=self.native_constraints)
        if not matcher.match_operation(pdl_op_val, pdl_op, xdsl_op):
            return

        parent = self.pdl_rewrite_op.parent_op()
        assert isinstance(parent, pdl.PatternOp)
        for constraint_op in parent.walk():
            if isinstance(constraint_op, pdl.ApplyNativeConstraintOp):
                if not matcher.check_native_constraints(constraint_op):
                    return

        self.interpreter.push_scope("rewrite")
        self.interpreter.set_values(matcher.matching_context.items())
        self.functions.rewriter = rewriter

        self.interpreter.run_ssacfg_region(self.pdl_rewrite_op.body, ())

        self.interpreter.pop_scope()


@register_impls
@dataclass
class PDLRewriteFunctions(InterpreterFunctions):
    """
    The implementations in this class are for the RHS of the rewrite. The SSA values
    referenced within the rewrite block are guaranteed to have been matched with the
    corresponding IR elements. The interpreter context stores the IR elements by SSA
    values.
    """

    ctx: Context
    _rewriter: PatternRewriter | None = field(default=None)

    @property
    def rewriter(self) -> PatternRewriter:
        assert self._rewriter is not None
        return self._rewriter

    @rewriter.setter
    def rewriter(self, rewriter: PatternRewriter):
        self._rewriter = rewriter

    @impl(pdl.OperationOp)
    def run_operation(
        self, interpreter: Interpreter, op: pdl.OperationOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        assert op.opName is not None
        op_name = op.opName.data
        op_type = self.ctx.get_optional_op(op_name)

        if op_type is None:
            raise InterpretationError(
                f"Could not find op type for name {op_name} in context"
            )

        attribute_value_names = [avn.data for avn in op.attributeValueNames.data]

        # How to deal with operandSegmentSizes?
        # operand_values, attribute_values, type_values = args

        operand_values = interpreter.get_values(op.operand_values)
        for operand in operand_values:
            assert isinstance(operand, SSAValue)

        attribute_values = interpreter.get_values(op.attribute_values)

        for attribute in attribute_values:
            assert isinstance(attribute, Attribute)

        type_values = interpreter.get_values(op.type_values)

        for type_value in type_values:
            assert isinstance(type_value, TypeAttribute)

        attributes = dict[str, Attribute]()
        properties = dict[str, Attribute]()

        # If the op is an IRDL-defined operation, get the property names.
        if issubclass(op_type, IRDLOperation):
            property_names = op_type.get_irdl_definition().properties.keys()
        else:
            property_names = []

        # Move the attributes to the attribute or property dictionary
        # depending on whether they are a properties or not.
        for attribute_name, attribute_value in zip(
            attribute_value_names, attribute_values
        ):
            if attribute_name in property_names:
                properties[attribute_name] = attribute_value
            else:
                attributes[attribute_name] = attribute_value

        result_op = op_type.create(
            operands=operand_values,
            result_types=type_values,
            attributes=attributes,
            properties=properties,
        )

        # Insert the new operation before the root operation
        self.rewriter.insert_op_before_matched_op(result_op)

        return (result_op,)

    @impl(pdl.ResultOp)
    def run_result(
        self, interpreter: Interpreter, op: pdl.ResultOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        (parent,) = args
        assert isinstance(parent, Operation)
        return (parent.results[op.index.value.data],)

    @impl(pdl.AttributeOp)
    def run_attribute(
        self, interpreter: Interpreter, op: pdl.AttributeOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        assert isinstance(op.value, Attribute)
        return (op.value,)

    @impl(pdl.ReplaceOp)
    def run_replace(
        self, interpreter: Interpreter, op: pdl.ReplaceOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        rewriter = self.rewriter

        (old,) = interpreter.get_values((op.op_value,))

        if op.repl_operation is not None:
            (new_op,) = interpreter.get_values((op.repl_operation,))
            rewriter.replace_op(old, new_ops=[], new_results=new_op.results)
        elif len(op.repl_values):
            new_vals = interpreter.get_values(op.repl_values)
            rewriter.replace_op(old, new_ops=[], new_results=list(new_vals))
        else:
            raise ValueError(
                "Either a replacing operatoin or values must be provided with replace op"
            )

        return ()

    @impl(pdl.EraseOp)
    def run_erase(
        self, interpreter: Interpreter, op: pdl.EraseOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        (old,) = interpreter.get_values((op.op_value,))
        self.rewriter.erase_op(old)
        return ()

    @impl(pdl.TypeOp)
    def run_type(
        self, interpreter: Interpreter, op: pdl.TypeOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        assert isinstance(op.constantType, Attribute)
        return (op.constantType,)



from __future__ import annotations

from xdsl.dialects import arith, builtin
from xdsl.ir import Dialect, Operation, SSAValue, VerifyException
from xdsl.irdl import (
    IRDLOperation,
    attr_def,
    irdl_op_definition,
    operand_def,
    var_operand_def,
)
from xdsl.parser import Parser
from xdsl.printer import Printer

i8 = builtin.IntegerType(8)


@irdl_op_definition
class PrintFormatOp(IRDLOperation):
    """
    A string formatting and printing utility.

    Can be though of as a printf equivalent but with python style format strings.

    Example uses:
    %42 = arith.constant 42 : i32
    printf.print_format "The magic number is {}", %42
    """

    name = "printf.print_format"

    format_str = attr_def(builtin.StringAttr)
    format_vals = var_operand_def()

    def __init__(self, format_str: str, *vals: SSAValue | Operation):
        super().__init__(
            operands=[vals], attributes={"format_str": builtin.StringAttr(format_str)}
        )

    def verify_(self) -> None:
        num_of_templates = self.format_str.data.count("{}")
        if not num_of_templates == len(self.format_vals):
            raise VerifyException(
                "Number of templates in template string must match number of arguments!"
            )

    def print(self, printer: Printer):
        printer.print_string(" ")
        printer.print_attribute(self.format_str)

        def print_val_and_type(ssa_val: SSAValue):
            printer.print_ssa_value(ssa_val)
            printer.print_string(" : ")
            printer.print_attribute(ssa_val.type)

        if len(self.format_vals) > 0:
            printer.print_string(", ")
            printer.print_list(self.operands, print_val_and_type)

        if len(self.attributes) > 1:
            attrs = self.attributes.copy()
            attrs.pop("format_str")
            printer.print_op_attributes(attrs)

    @classmethod
    def parse(cls: type[PrintFormatOp], parser: Parser) -> PrintFormatOp:
        format_str = parser.parse_str_literal()
        args: list[SSAValue] = []
        while parser.parse_optional_characters(",") is not None:
            args.append(arg := parser.parse_operand())
            parser.parse_characters(":", " - all arguments must have a type")
            arg_type = parser.parse_type()
            if arg.type != arg_type:
                parser.raise_error(f"Parsed ssa vlue {arg} must be of type {arg_type}")

        attr_dict = parser.parse_optional_attr_dict()

        if "format_str" in attr_dict:
            parser.raise_error(
                "format_str keyword is a reserved attribute for printf.print_format!"
            )

        op = PrintFormatOp(format_str, *args)

        op.attributes.update(attr_dict)

        return op


@irdl_op_definition
class PrintCharOp(IRDLOperation):
    """
    Print a single character

    Equivalent to putchar in C, but uses signless bytes as input (instead of ui32).
    Unlike the C implementation, this op does not return anything.
    """

    name = "printf.print_char"
    char = operand_def(i8)

    def __init__(self, char: SSAValue | Operation):
        super().__init__(
            operands=[char],
        )

    @staticmethod
    def from_constant_char(char: str) -> PrintCharOp:
        """
        This constructor returns a PrintCharOp that prints the value supplied
        in "char" as a python char.
        """
        if len(char) != 1:
            raise ValueError(
                f'Unexpected char value "{char}", input must be a single ascii character'
            )
        ascii_value = ord(char)
        if ascii_value > 128:
            raise ValueError("Only ascii characters are supported")
        char_constant = arith.ConstantOp.from_int_and_width(ascii_value, i8)
        return PrintCharOp(char_constant)


@irdl_op_definition
class PrintIntOp(IRDLOperation):
    """
    Print a single Integer
    """

    name = "printf.print_int"
    int = operand_def(builtin.IntegerType)

    def __init__(self, integer: SSAValue | Operation):
        super().__init__(
            operands=[integer],
        )


Printf = Dialect(
    "printf",
    [
        PrintFormatOp,
        PrintCharOp,
        PrintIntOp,
    ],
    [],
)



import pytest

from xdsl.dialects import tensor
from xdsl.dialects.builtin import ModuleOp, TensorType, f32, i32
from xdsl.interpreter import Interpreter
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.tensor import TensorFunctions
from xdsl.interpreters.utils.ptr import TypedPtr
from xdsl.utils.exceptions import InterpretationError
from xdsl.utils.test_value import create_ssa_value


def test_tensor_empty():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(TensorFunctions())
    op = tensor.EmptyOp((), TensorType(f32, [3, 2]))
    (c,) = interpreter.run_op(op, ())
    assert c == ShapedArray(TypedPtr.new_float32((0,) * 6), [3, 2])


def test_tensor_reshape():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(TensorFunctions())
    op = tensor.ReshapeOp(
        create_ssa_value(TensorType(f32, [4, 1])),
        create_ssa_value(TensorType(i32, [1])),
        TensorType(f32, [4]),
    )
    a = ShapedArray(TypedPtr.new_float32([1, 2, 3, 4]), [4, 1])
    b = ShapedArray(TypedPtr.new_float32([4]), [1])
    (c,) = interpreter.run_op(op, (a, b))
    assert c == ShapedArray(TypedPtr.new_float32([1, 2, 3, 4]), [4])


def test_tensor_reshape_error():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(TensorFunctions())
    op = tensor.ReshapeOp(
        create_ssa_value(TensorType(f32, [3, 1])),
        create_ssa_value(TensorType(i32, [1])),
        TensorType(f32, [3]),
    )
    a = ShapedArray(TypedPtr.new_float32([1, 2, 3]), [3, 1])
    b = ShapedArray(TypedPtr.new_float32([2]), [1])
    with pytest.raises(
        InterpretationError, match="Mismatch between static shape and new shape"
    ):
        interpreter.run_op(op, (a, b))



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "xdsl==0.27.0",
# ]
# ///

import marimo

__generated_with = "0.14.11"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo
    from sympy import (
        S,
        symbols,
        Expr,
        Add,
        Mul,
        Sum,
        Integer,
        Float,
        E,
        I,
        re,
        im,
        Abs,
        Pow,
        Rational,
        Function,
        UnevaluatedExpr,
    )
    from sympy.core.symbol import Symbol

    from xdsl.ir import Attribute, Operation, SSAValue, Region, Block, ParametrizedAttribute
    from xdsl.pattern_rewriter import (
        PatternRewriter,
        RewritePattern,
        op_type_rewrite_pattern,
        PatternRewriteWalker,
        GreedyRewritePatternApplier,
    )
    from xdsl.transforms.dead_code_elimination import region_dce
    from xdsl.traits import Pure
    from xdsl.irdl import (
        irdl_op_definition,
        traits_def,
        IRDLOperation,
        irdl_attr_definition,
        operand_def,
        result_def,
    )
    from xdsl.dialects.builtin import (
        ModuleOp,
        Float64Type,
        FloatAttr,
        IntegerType,
        IntegerAttr,
    )
    from xdsl.dialects.func import FuncOp, ReturnOp
    from xdsl.dialects.arith import (
        AddfOp,
        SubfOp,
        MulfOp,
        ConstantOp,
        AddiOp,
        MuliOp,
        SIToFPOp,
        FloatingPointLikeBinaryOperation,
        DivfOp,
    )
    from xdsl.dialects.scf import ForOp, YieldOp
    from xdsl.dialects.math import PowFOp, SqrtOp
    from xdsl.builder import Builder, InsertPoint
    return (
        Abs,
        Add,
        AddfOp,
        AddiOp,
        Attribute,
        Block,
        Builder,
        ConstantOp,
        Expr,
        Float,
        Float64Type,
        FloatAttr,
        ForOp,
        FuncOp,
        GreedyRewritePatternApplier,
        InsertPoint,
        Integer,
        IntegerAttr,
        IntegerType,
        ModuleOp,
        Mul,
        MulfOp,
        MuliOp,
        Operation,
        PatternRewriteWalker,
        PatternRewriter,
        Pow,
        PowFOp,
        Region,
        ReturnOp,
        RewritePattern,
        SIToFPOp,
        SSAValue,
        SubfOp,
        Sum,
        Symbol,
        YieldOp,
        mo,
        region_dce,
        symbols,
    )


@app.cell(hide_code=True)
def _(Expr, emit_ir):
    def print_ir(expr: Expr):
        # Print the SymPy expression
        print(expr)

        # Converts the SymPy expression to an MLIR `builtin.module` operation
        op = emit_ir(expr)

        # Check that the operation verifies, and prints the operation
        op.verify()
        print(op)

        # Print a separator
        print("\n\n")
    return (print_ir,)


@app.cell(hide_code=True)
def _(Attribute, Expr, Float64Type, IntegerType):
    # Get the MLIR type for a SymPy expression
    def get_mlir_type(expr: Expr) -> Attribute:
        if expr.is_integer:
            return IntegerType(64)
        elif expr.is_extended_real:
            return Float64Type()
        else:
            raise Exception(
                f"Unknown MLIR type for expression {expr}. Please make sure there cannot be a division by zero, or a power of a negative value."
            )
    return (get_mlir_type,)


@app.cell(hide_code=True)
def _(
    Builder,
    Expr,
    FuncOp,
    InsertPoint,
    ModuleOp,
    ReturnOp,
    emit_op,
    get_mlir_type,
):
    def emit_ir(expr: Expr) -> ModuleOp:
        # Create a module, and create a builder at the beginning of its only block
        module = ModuleOp([])
        builder = Builder(InsertPoint.at_end(module.body.block))

        # Create the MLIR types for each symbol.
        arg_types = [get_mlir_type(arg) for arg in expr.free_symbols]

        # Create a new function and inserts it inside the module.
        func = FuncOp("main", (arg_types, [get_mlir_type(expr)]))
        builder.insert(func)

        # Associate each symbol with its MLIR name.
        arg_values = {arg: value for arg, value in zip(expr.free_symbols, func.args)}

        # Set the name for each function argument. This is only to get better names
        # for IR values.
        for arg, value in arg_values.items():
            value.name_hint = arg.name

        # Set the builder insertion point inside the function.
        builder.insertion_point = InsertPoint.at_end(func.body.block)

        # Convert the expression into MLIR IR inside the function.
        result = emit_op(expr, builder, arg_values)

        # Insert a return statement at the end of the function.
        builder.insert(ReturnOp(result))
        return module
    return (emit_ir,)


@app.cell(hide_code=True)
def _(
    Builder,
    Expr,
    Float64Type,
    IntegerType,
    SSAValue,
    Symbol,
    emit_integer_op,
    emit_real_op,
    get_mlir_type,
):
    def emit_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        type = get_mlir_type(expr)
        if isinstance(type, IntegerType):
            return emit_integer_op(expr, builder, args)
        elif isinstance(type, Float64Type):
            return emit_real_op(expr, builder, args)
        else:
            raise Exception("Unknown function to emit IR for MLIR type ", type)
    return (emit_op,)


@app.cell(hide_code=True)
def _(
    Abs,
    Add,
    AddfOp,
    AddiOp,
    Block,
    Builder,
    CmpfOp,
    ConstantOp,
    Expr,
    Float,
    Float64Type,
    FloatAttr,
    ForOp,
    IfOp,
    InsertPoint,
    Integer,
    IntegerAttr,
    IntegerType,
    Mul,
    MulfOp,
    MuliOp,
    Pow,
    PowFOp,
    Region,
    SIToFPOp,
    SSAValue,
    SelectOp,
    SubfOp,
    Sum,
    Symbol,
    YieldOp,
):
    def emit_integer_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        # Handle symbolic values
        if isinstance(expr, Symbol):
            return args[expr]

        # Handle constants
        if isinstance(expr, Integer):
            constant_op = builder.insert(
                ConstantOp(IntegerAttr(int(expr), IntegerType(64)))
            )
            return constant_op.result

        if isinstance(expr, Add):
            lhs = emit_integer_op(expr.args[0], builder, args)
            rhs = emit_integer_op(expr.args[1], builder, args)
            add_op = builder.insert(AddiOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Mul):
            lhs = emit_integer_op(expr.args[0], builder, args)
            rhs = emit_integer_op(expr.args[1], builder, args)
            add_op = builder.insert(MuliOp(lhs, rhs))
            return add_op.result

        # Hint: Implement here support for Add and Mul

        raise NotImplementedError(f"No IR emitter for integer function {expr.func}")

    def emit_real_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        # If the expression is an integer expression, emits it and then convert it
        # back to a float expression.
        if expr.is_integer:
            res = emit_integer_op(expr, builder, args)
            op = builder.insert(SIToFPOp(res, Float64Type()))
            return op.result

        # Handle constants
        if isinstance(expr, Float):
            constant_op = builder.insert(
                ConstantOp(FloatAttr(float(expr), Float64Type()))
            )
            return constant_op

        # Handle symbolic values
        if isinstance(expr, Symbol):
            return args[expr]

        if isinstance(expr, Add):
            lhs = emit_real_op(expr.args[0], builder, args)
            rhs = emit_real_op(expr.args[1], builder, args)
            add_op = builder.insert(AddfOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Mul):
            lhs = emit_real_op(expr.args[0], builder, args)
            rhs = emit_real_op(expr.args[1], builder, args)
            add_op = builder.insert(MulfOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Pow):
            lhs = emit_real_op(expr.args[0], builder, args)
            rhs = emit_real_op(expr.args[1], builder, args)
            add_op = builder.insert(PowFOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Abs):
            # The arith.select solution
            if False:
                arg = emit_real_op(expr.args[0], builder, args)
                zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
                neg = builder.insert(SubfOp(zero, arg)).result
                is_neg = builder.insert(CmpfOp(arg, zero, "olt")).result
                select = builder.insert(SelectOp(is_neg, neg, arg)).result
                return select

            # The scf.if solution
            arg = emit_real_op(expr.args[0], builder, args)
            zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
            is_neg = builder.insert(CmpfOp(arg, zero, "olt")).result

            lhs_region = Region([Block()])
            builder2 = Builder(InsertPoint.at_end(lhs_region.block))
            neg = builder2.insert(SubfOp(zero, arg)).result
            builder2.insert(YieldOp(neg))

            rhs_region = Region([Block()])
            builder3 = Builder(InsertPoint.at_end(rhs_region.block))
            builder3.insert(YieldOp(neg))

            if_res = builder.insert(
                IfOp(is_neg, Float64Type(), lhs_region, rhs_region)
            ).results[0]

            return if_res

        if isinstance(expr, Sum):
            zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
            lb = emit_integer_op(expr.args[1][1], builder, args)
            ub = emit_integer_op(expr.args[1][2], builder, args)
            step = builder.insert(ConstantOp(IntegerAttr(0, IntegerType(64))))
            region = Region([Block(arg_types=[IntegerType(64), Float64Type()])])
            accumulator = region.block.args[1]

            b2 = Builder(InsertPoint.at_end(region.block))
            arg = emit_real_op(
                expr.args[0], b2, args | {expr.args[1][0]: region.block.args[0]}
            )
            add = b2.insert(AddfOp(arg, accumulator)).result
            b2.insert(YieldOp(add))

            return builder.insert(ForOp(lb, ub, step, [zero], region)).results[0]

        # Hint: Implement here support for Add, Mul, and Pow (and later Abs and Sum)

        raise NotImplementedError(f"No IR emitter for float function {expr.func}")
    return emit_integer_op, emit_real_op


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
    # Pattern Rewrite Exercises

    In the previous exercise, the code we emitted had a few optimization opportunities.
    We task you here to write two of them here, to get a feeling how rewrites work in xDSL.

    The objective is to improve the code generated by `x - y`:
    """
    )
    return


@app.cell
def _(print_ir, symbols):
    # x, y, z, t are non-zere reals.
    x, y, z, t = symbols("x y z t", real=True, zero=False)

    print_ir(x - y)
    return x, y


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
    As you can see, SymPy "simplifies" the expression to be `x + -1 * y`, which in turns gets transformed to `x + (float)(-1) * y`.
    While we could match for this exact pattern, defining two smaller rewrite patterns would be both simpler and allow more reuse.

    So, we task you to write the following optimizations:

    * `(float)(int_constant) -> float_constant`, which will convert `(float)(-1)` to `-1.0`
    * `x + -1.0 * y -> x - y`, which will be applied after application of the first rewrite

    Here is the following boilerplate that you should complete:
    """
    )
    return


@app.cell
def _(Operation, PatternRewriter, RewritePattern):
    class SIToFPConstantPattern(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            # Implement `(float)(int_constant) -> float_constant` here
            return

    class AddTimesMinusOnePattern(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            # Implement `x + -1.0 * y -> x - y` here
            return
    return AddTimesMinusOnePattern, SIToFPConstantPattern


@app.cell(hide_code=True)
def _(mo):
    mo.md("""You can run the optimizations on a SymPy expression in the following code blocks:""")
    return


@app.cell(hide_code=True)
def _(
    AddTimesMinusOnePattern,
    Expr,
    GreedyRewritePatternApplier,
    PatternRewriteWalker,
    SIToFPConstantPattern,
    emit_ir,
    region_dce,
):
    def print_with_opts(expr: Expr):
        print(expr)
        op = emit_ir(expr)
        op.verify()
        print("Before optimizations:", op)
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [SIToFPConstantPattern(), AddTimesMinusOnePattern()]
            )
        ).rewrite_module(op)
        region_dce(op.body)
        print("After optimizations:", op)
        print("\n" * 3)
    return (print_with_opts,)


@app.cell
def _(print_with_opts, x, y):
    print_with_opts(x - y)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
    You should get the resulting IR:

    ```
    builtin.module {
      func.func @main(%y : f64, %x : f64) -> f64 {
        %0 = arith.subf %x, %y : f64
        func.return %0 : f64
      }
    }
    ```
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    # Solutions

    Here is a solution for both rewrites:
    """
    )
    return


@app.cell(hide_code=True)
def _(
    AddfOp,
    ConstantOp,
    Float64Type,
    FloatAttr,
    MulfOp,
    Operation,
    PatternRewriter,
    RewritePattern,
    SIToFPOp,
    SubfOp,
):
    class SIToFPConstantPatternSolution(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if not isinstance(op, SIToFPOp):
                return
            if not isinstance(op.input.owner, ConstantOp):
                return
            new_op = ConstantOp(
                FloatAttr(op.input.owner.value.value.data, Float64Type())
            )
            rewriter.replace_op(op, new_op)

    class AddTimesMinusOnePatternSolution(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if not isinstance(op, AddfOp):
                return
            if not isinstance(mul := op.rhs.owner, MulfOp):
                return
            if not isinstance(constant := mul.lhs.owner, ConstantOp):
                return
            if constant.value.value.data != -1.0:
                return
            rewriter.replace_op(op, SubfOp(op.lhs, mul.rhs))
    return


if __name__ == "__main__":
    app.run()



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, SupportsIndex, cast, overload

from typing_extensions import Self, TypeVar

_T = TypeVar("_T")
_S = TypeVar("_S")


class IList(list[_T]):
    """
    A list that can be frozen. Once frozen, it can not be modified.
    In comparison to FrozenList this supports pattern matching.
    """

    _frozen: bool = False

    def freeze(self):
        self._frozen = True

    def _unfreeze(self):
        self._frozen = False

    def append(self, object: _T, /) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().append(object)

    def extend(self, iterable: Iterable[_T], /) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().extend(iterable)

    def insert(self, index: SupportsIndex, object: _T, /) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().insert(index, object)

    def remove(self, value: _T, /) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().remove(value)

    def pop(self, index: SupportsIndex = -1, /) -> _T:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().pop(index)

    def clear(self) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().clear()

    @overload
    def __setitem__(self, key: SupportsIndex, value: _T) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[_T]) -> None: ...

    def __setitem__(self, key: SupportsIndex | slice, value: _T | Iterable[_T]) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        if isinstance(key, slice):
            o = cast(list[_T], value)
            return super().__setitem__(key, o)
        else:
            o = cast(_T, value)
            return super().__setitem__(key, o)

    def __delitem__(self, key: SupportsIndex | slice, /) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().__delitem__(key)

    def __add__(self, value: Iterable[_S]) -> IList[_T | _S]:
        return IList(super().__add__(list(value)))

    def __iadd__(self, value: Iterable[_T], /) -> Self:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().__iadd__(value)

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, IList):
            other = cast(IList[Any], value)
            return super().__eq__(other)
        return False



"""
This file contains the definition of `BaseParser`, a recursive descent parser
that is inherited from the different parsers used in xDSL.
"""

from collections.abc import Callable, Iterable
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Generic, NoReturn, overload

from typing_extensions import TypeVar

from xdsl.utils.exceptions import ParseError
from xdsl.utils.lexer import Lexer, Position, Span, Token, TokenKindT


@dataclass(init=False)
class ParserState(Generic[TokenKindT]):
    """
    The parser state. It contains the lexer, and the next token to parse.
    The parser state should be shared between all parsers, so parsers can
    share the same position.
    """

    lexer: Lexer[TokenKindT]
    current_token: Token[TokenKindT]
    dialect_stack: list[str]

    def __init__(
        self, lexer: Lexer[TokenKindT], dialect_stack: list[str] | None = None
    ):
        if dialect_stack is None:
            dialect_stack = ["builtin"]
        self.lexer = lexer
        self.current_token = lexer.lex()
        self.dialect_stack = dialect_stack


_AnyInvT = TypeVar("_AnyInvT")


@dataclass
class GenericParser(Generic[TokenKindT]):
    """
    Basic recursive descent parser. It contains parsing methods that are shared
    between the different parsers.

    Methods named `parse_*` will consume tokens, and throw a `ParseError` if
    an unexpected token is parsed. Methods marked with `parse_optional` will return
    None if the first token is unexpected, and will throw a `ParseError` if the
    first token is expected, but a following token is not.

    Methods with a `context_msg` argument allows to append the context message to the
    thrown error. For instance, if `',' expected` is returned, setting `context_msg` to
    `" in integer list"` will throw the error `',' expected in integer list` instead.
    """

    _parser_state: ParserState[TokenKindT]

    @overload
    def raise_error(
        self,
        msg: str,
        at_position: Position,
        end_position: Position,
    ) -> NoReturn: ...

    @overload
    def raise_error(
        self,
        msg: str,
        at_position: Position | Span | None = None,
    ) -> NoReturn: ...

    def raise_error(
        self,
        msg: str,
        at_position: Span | Position | None = None,
        end_position: Position | None = None,
    ) -> NoReturn:
        """
        Raise a parsing exception at a specific location.
        If no location is provided, the current parser state location is used.
        """
        if end_position is not None:
            assert isinstance(at_position, Position)
            at_position = Span(at_position, end_position, self.lexer.input)
        if at_position is None:
            at_position = self._current_token.span
        elif isinstance(at_position, Position):
            at_position = Span(at_position, at_position, self.lexer.input)

        raise ParseError(at_position, msg)

    def _resume_from(self, pos: Position | ParserState[TokenKindT]):
        """Resume parsing from a given position."""
        if isinstance(pos, Position):
            self.lexer.pos = pos
            self._parser_state.current_token = self.lexer.lex()
        else:
            self._parser_state = pos

    @property
    def _current_token(self) -> Token[TokenKindT]:
        """Get the token that is currently being parsed. Do not consume the token."""
        return self._parser_state.current_token

    @property
    def lexer(self) -> Lexer[TokenKindT]:
        """The lexer used to parse the current input."""
        return self._parser_state.lexer

    @property
    def pos(self) -> Position:
        """
        Get the starting position of the next token.
        This skips the whitespaces.
        """
        return self._current_token.span.start

    def _consume_token(
        self, expected_kind: TokenKindT | None = None
    ) -> Token[TokenKindT]:
        """
        Advance the lexer to the next token.
        Assert that the current token was of a specific kind, if specified.
        For reporting errors if the token was not of the expected kind,
        use `_parse_token` instead.
        """
        consumed_token = self._current_token
        assert expected_kind is None or consumed_token.kind == expected_kind, (
            f"Unexpected token {consumed_token}, expected {expected_kind}"
        )
        self._parser_state.current_token = self.lexer.lex()
        return consumed_token

    def _parse_optional_token(
        self, expected_kind: TokenKindT
    ) -> Token[TokenKindT] | None:
        """
        If the current token is of the expected kind, consume it and return it.
        Otherwise, return None.
        """
        if self._current_token.kind == expected_kind:
            return self._consume_token()

    def _parse_token(
        self, expected_kind: TokenKindT, error_msg: str
    ) -> Token[TokenKindT]:
        """
        Parse a specific token, and raise an error if it is not present.
        Returns the token that was parsed.
        """
        if (result := self._parse_optional_token(expected_kind)) is None:
            self.raise_error(error_msg, self._current_token.span)
        return result

    def _parse_optional_token_in(
        self, expected_kinds: Iterable[TokenKindT]
    ) -> Token[TokenKindT] | None:
        """Parse one of the expected tokens if present, and returns it."""
        if self._current_token.kind not in expected_kinds:
            return None
        return self._consume_token()

    def expect(
        self, try_parse: Callable[[], _AnyInvT | None], error_message: str
    ) -> _AnyInvT:
        """
        This method is used to convert a `parse_optional_*` to a `parse_*`.
        It will run the parsing function, and raise an error if `None` was returned.
        """
        res = try_parse()
        if res is None:
            self.raise_error(error_message)
        return res

    class Delimiter(Enum):
        """
        Supported delimiters when parsing lists.
        """

        PAREN = ("(", ")")
        ANGLE = ("<", ">")
        SQUARE = ("[", "]")
        BRACES = ("{", "}")
        METADATA_TOKEN = ("{-#", "#-}")
        NONE = None

    def parse_comma_separated_list(
        self, delimiter: Delimiter, parse: Callable[[], _AnyInvT], context_msg: str = ""
    ) -> list[_AnyInvT]:
        """
        Parses greedily a list of elements separated by commas, and delimited
        by the specified delimiter. The parsing stops when the delimiter is
        closed, or when an error is produced. If no delimiter is specified, at
        least one element is expected to be parsed.
        """
        # Parse the opening bracket, and possibly the closing bracket, if a delimiter
        # was provided
        match delimiter.value:
            case None:
                pass
            case (left_punctuation, right_punctuation):
                self.parse_characters(left_punctuation, context_msg)
                if self.parse_optional_characters(right_punctuation) is not None:
                    return []

        # Parse the list of elements
        elems = [parse()]
        while self.parse_optional_characters(",") is not None:
            elems.append(parse())

        # Parse the closing bracket, if a delimiter was provided
        match delimiter.value:
            case None:
                pass
            case (_, right_punctuation):
                self.parse_characters(right_punctuation, context_msg)

        return elems

    def parse_optional_comma_separated_list(
        self, delimiter: Delimiter, parse: Callable[[], _AnyInvT], context_msg: str = ""
    ) -> list[_AnyInvT] | None:
        """
        Parses greedily a list of elements separated by commas, and delimited
        by the specified delimiter. If no opening delimiter was found, return None.
        The parsing stops when the delimiter is closed, or when an error is produced.
        The NONE delimiter is not allowed by this method, use
        `parse_optional_undelimited_comma_separated_list` instead.
        """

        if delimiter == self.Delimiter.NONE:
            raise ValueError(
                "Cannot use `Delimiter.NONE` with "
                "`parse_optional_comma_separated_list`. Use "
                "`parse_optional_undelimited_comma_separated_list` instead."
            )

        # Parse the opening bracket, and possibly the closing bracket
        left_punctuation, right_punctuation = delimiter.value
        if self.parse_optional_characters(left_punctuation) is None:
            return None
        if self.parse_optional_characters(right_punctuation) is not None:
            return []

        # Parse the list of elements
        elems = [parse()]
        while self.parse_optional_characters(",") is not None:
            elems.append(parse())

        # Parse the closing bracket
        self.parse_characters(right_punctuation, context_msg)

        return elems

    def parse_optional_undelimited_comma_separated_list(
        self,
        parse_optional: Callable[[], _AnyInvT | None],
        parse: Callable[[], _AnyInvT],
    ) -> list[_AnyInvT] | None:
        """
        Parses greedily a list of elements separated by commas, if a first element is
        present. The first element is parsed with `parse_optional`, and the remaining
        are parsed with `parse`. The parsing stops either if the first element is not
        present, or if no comma is present after parsing an element.
        """
        # Parse the first element, if it exist
        first_elem = parse_optional()
        if first_elem is None:
            return None

        # Parse the remaining elements
        elems = [first_elem]
        while self.parse_optional_characters(",") is not None:
            elems.append(parse())

        return elems

    def parse_optional_characters(self, text: str) -> str | None:
        """
        Parse a given token text, if present.
        If the given text is the beginning of the next token, this will still
        return None.
        """
        if self._current_token.text == text:
            self._consume_token()
            return text
        return None

    def parse_characters(self, text: str, context_msg: str = "") -> str:
        """
        Parse a given token text.
        The context message is appended to the error message if the parsing fails.
        If the given text is the start of the next token, this will still raise
        an error.
        """
        if (res := self.parse_optional_characters(text)) is not None:
            return res
        self.raise_error(f"'{text}' expected" + context_msg)

    @contextmanager
    def delimited(self, start: str, end: str):
        self.parse_characters(start)
        yield
        self.parse_characters(end)

    def in_angle_brackets(self):
        return self.delimited("<", ">")

    def in_square_brackets(self):
        return self.delimited("[", "]")

    def in_parens(self):
        return self.delimited("(", ")")

    def in_braces(self):
        return self.delimited("{", "}")



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic

from typing_extensions import TypeVar

from xdsl.dialects.builtin import (
    I32,
    I64,
    IndexType,
    IntegerAttr,
    IntegerType,
    i1,
    i32,
    i64,
)
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    AttrSizedOperandSegments,
    IRDLOperation,
    irdl_op_definition,
    operand_def,
    prop_def,
    result_def,
    traits_def,
    var_operand_def,
)
from xdsl.traits import NoMemoryEffect
from xdsl.utils.exceptions import VerifyException

# Transfer ID
tx_id = i32
# Indicates address range in memory, a "memory slice"
slice_t_begin = i64
slice_t_end = i64


class SnitchRuntimeBaseOperation(IRDLOperation, ABC):
    """
    A base class for ops in the Snitch Runtime dialect.
    The Snitch Runtime dialect models the snitch runtime which contains low level
    routines to manage system level aspects of snitch systems.

    This dialect is modeled after:
    https://github.com/pulp-platform/snitch_cluster/tree/main/sw/snRuntime
    """

    pass


class SnitchRuntimeGetInfo(SnitchRuntimeBaseOperation, ABC):
    """
    A base class for snitch runtime functions that get a certain value at runtime
    """

    result = result_def(i32)

    traits = traits_def(NoMemoryEffect())

    def __init__(
        self,
    ):
        super().__init__(result_types=[i32])


class SnitchRuntimeGetInfoBool(SnitchRuntimeBaseOperation, ABC):
    """
    A base class for snitch runtime functions that get a certain value at runtime
    """

    result = result_def(i1)

    traits = traits_def(NoMemoryEffect())

    def __init__(
        self,
    ):
        super().__init__(result_types=[i1])


class NoOperandNoResultBaseOperation(SnitchRuntimeBaseOperation, ABC):
    """
    A base class for operations with no operands nor results
    """

    def __init__(
        self,
    ):
        super().__init__()


@irdl_op_definition
class GlobalCoreBaseHartidOp(SnitchRuntimeGetInfo):
    """
    Get the current core's global base Hart ID
    """

    name = "snrt.global_core_base_hartid"


@irdl_op_definition
class GlobalCoreIdxOp(SnitchRuntimeGetInfo):
    """
    Regardless of core type, return global core index, equal to the Hart ID of the
    current core - global base Hart ID of the cluster
    """

    name = "snrt.global_core_idx"


@irdl_op_definition
class GlobalCoreNumOp(SnitchRuntimeGetInfo):
    """
    Return total amount of cores including DMA cores per cluster
    """

    name = "snrt.global_core_num"


@irdl_op_definition
class GlobalComputeCoreIdxOp(SnitchRuntimeGetInfo):
    """
    For compute core, return global core index
    """

    name = "snrt.global_compute_core_idx"


@irdl_op_definition
class GlobalComputeCoreNumOp(SnitchRuntimeGetInfo):
    """
    Return total amount of compute cores per cluster
    """

    name = "snrt.global_compute_core_num"


@irdl_op_definition
class GlobalDmCoreNumOp(SnitchRuntimeGetInfo):
    """
    Return total amount of DMA cores
    """

    name = "snrt.global_dm_core_num"


@irdl_op_definition
class ClusterCoreIdxOp(SnitchRuntimeGetInfo):
    """
    Return cluster identifier
    """

    name = "snrt.cluster_core_idx"


@irdl_op_definition
class ClusterCoreNumOp(SnitchRuntimeGetInfo):
    """
    Return total amount of cores for the current cluster
    """

    name = "snrt.cluster_core_num"


@irdl_op_definition
class ClusterComputeCoreIdxOp(SnitchRuntimeGetInfo):
    """
    For compute cores return core ID within a cluster
    """

    name = "snrt.cluster_compute_core_idx"


@irdl_op_definition
class ClusterComputeCoreNumOp(SnitchRuntimeGetInfo):
    """
    Return number of compute cores for the current cluster
    """

    name = "snrt.cluster_compute_core_num"


@irdl_op_definition
class ClusterDmCoreIdxOp(SnitchRuntimeGetInfo):
    """
    For DMA cores, return core ID within a cluster, currently hardcoded to number of all cores - 1
    """

    name = "snrt.cluster_dm_core_idx"


@irdl_op_definition
class ClusterDmCoreNumOp(SnitchRuntimeGetInfo):
    """
    Return amount of DMA cores in this cluster, in the current runtime, this is hardcoded to 1
    """

    name = "snrt.cluster_dm_core_num"


@irdl_op_definition
class ClusterIdxOp(SnitchRuntimeGetInfo):
    """
    Return i32 identifier for the cluster this core is a part of
    """

    name = "snrt.cluster_idx"


@irdl_op_definition
class ClusterNumOp(SnitchRuntimeGetInfo):
    """
    Probe the amount of clusters
    """

    name = "snrt.cluster_num"


@irdl_op_definition
class IsComputeCoreOp(SnitchRuntimeGetInfoBool):
    """
    Return non-zero integer if current snitch core is a compute core
    """

    name = "snrt.is_compute_core"


@irdl_op_definition
class IsDmCoreOp(SnitchRuntimeGetInfoBool):
    """
    Return non-zero integer if current snitch core is a DMA core
    """

    name = "snrt.is_dm_core"


@irdl_op_definition
class ClusterHwBarrierOp(NoOperandNoResultBaseOperation):
    """
    Synchronize cores in a cluster with a hardware barrier
    """

    name = "snrt.cluster_hw_barrier"


@irdl_op_definition
class ClusterSwBarrierOp(NoOperandNoResultBaseOperation):
    """
    Synchronize with compute cores after loading data
    """

    name = "snrt.cluster_sw_barrier"


@irdl_op_definition
class GlobalBarrierOp(NoOperandNoResultBaseOperation):
    """
    Synchronize clusters globally with a global software barrier
    """

    name = "snrt.global_barrier"


_T = TypeVar("_T", bound=Attribute)


@irdl_op_definition
class BarrierRegPtrOp(SnitchRuntimeGetInfo):
    """
    Get pointer to barrier register
    """

    name = "snrt.barrier_reg_ptr"


class GetMemoryInfoBaseOperation(SnitchRuntimeBaseOperation, ABC):
    """
    Generic base class for operations returning memory slices
    """

    slice_begin = result_def(slice_t_begin)
    slice_end = result_def(slice_t_end)

    traits = traits_def(NoMemoryEffect())

    def __init__(
        self,
    ):
        super().__init__(result_types=[slice_t_begin, slice_t_end])


@irdl_op_definition
class GlobalMemoryOp(GetMemoryInfoBaseOperation):
    """
    Get start address of global memory
    """

    name = "snrt.global_memory"


@irdl_op_definition
class ClusterMemoryOp(GetMemoryInfoBaseOperation):
    """
    Get start address of the cluster's TCDM memory
    """

    name = "snrt.cluster_memory"


@irdl_op_definition
class ZeroMemoryOp(GetMemoryInfoBaseOperation):
    """
    Get start address of the cluster's zero memory
    """

    name = "snrt.zero_memory"


class DmaStart1DBaseOperation(SnitchRuntimeBaseOperation, Generic[_T], ABC):
    """
    Initiate an asynchronous 1D DMA transfer
    """

    dst = operand_def(_T)
    src = operand_def(_T)

    size = operand_def(i32)
    transfer_id = result_def(tx_id)

    def __init__(
        self,
        dst: Operation | SSAValue,
        src: Operation | SSAValue,
        size: Operation | SSAValue,
    ):
        super().__init__(operands=[dst, src, size], result_types=[tx_id])


class DmaStart2DBaseOperation(SnitchRuntimeBaseOperation, Generic[_T], ABC):
    """
    Generic base class for starting asynchronous 2D DMA transfers
    """

    dst = operand_def(_T)
    src = operand_def(_T)
    dst_stride = operand_def(i32)
    src_stride = operand_def(i32)
    size = operand_def(i32)
    repeat = operand_def(i32)
    transfer_id = result_def(tx_id)

    def __init__(
        self,
        dst: Operation | SSAValue,
        src: Operation | SSAValue,
        dst_stride: Operation | SSAValue,
        src_stride: Operation | SSAValue,
        size: Operation | SSAValue,
        repeat: Operation | SSAValue,
    ):
        super().__init__(
            operands=[dst, src, dst_stride, src_stride, size, repeat],
            result_types=[tx_id],
        )


@irdl_op_definition
class DmaStart1DOp(DmaStart1DBaseOperation[I32]):
    """
    Initiate an asynchronous 1D DMA transfer with 32-bits pointers
    """

    name = "snrt.dma_start_1d"


@irdl_op_definition
class DmaStart1DWideptrOp(DmaStart1DBaseOperation[I64]):
    """
    Initiate an asynchronous 1D DMA transfer with 64-bits wide pointers
    """

    name = "snrt.dma_start_1d_wideptr"


@irdl_op_definition
class DmaStart2DOp(DmaStart2DBaseOperation[I32]):
    """
    Initiate an asynchronous 2D DMA transfer with 32-bits pointers
    """

    name = "snrt.dma_start_2d"


@irdl_op_definition
class DmaStart2DWideptrOp(DmaStart2DBaseOperation[I64]):
    """
    Initiate an asynchronous 2D DMA transfer with 64-bits wide pointers
    """

    name = "snrt.dma_start_2d_wideptr"


@irdl_op_definition
class DmaWaitOp(SnitchRuntimeBaseOperation):
    """
    Block until a transfer finishes
    """

    name = "snrt.dma_wait"
    transfer_id = operand_def(tx_id)

    def __init__(self, transfer_id: Operation | SSAValue):
        super().__init__(operands=[transfer_id])


@irdl_op_definition
class DmaWaitAllOp(NoOperandNoResultBaseOperation):
    """
    Block until all operations on the DMA cease
    """

    name = "snrt.dma_wait_all"


class SsrLoopBaseOp(SnitchRuntimeBaseOperation, ABC):
    """
    Configure an SSR data mover for an n-dimensional loop nest.
    bounds (limits of loop) and strides (increments of size) are ordered from inner-most to outer-most loops.

    for example:
    for (i = 0; i < 5; i++) { //bounds[1] = 5 and strides[1] = 1
        for (j = 0; j < 6; j+=2) { //bounds[0] = 6 and strides[0] = 2
            ...
        }
    }
    """

    data_mover = operand_def(i32)
    bounds = var_operand_def(IndexType)
    strides = var_operand_def(IndexType)
    irdl_options = [AttrSizedOperandSegments()]

    def verify_(self) -> None:
        if len(self.bounds) != len(self.strides):
            raise VerifyException(
                f"the length of bounds ({len(self.bounds)}) and strides ({len(self.strides)}) must be equal."
            )
        if len(self.strides) != self.num:
            raise VerifyException(
                f"Epected {self.num} bounds and strides, got {len(self.strides)}"
            )

    def __init__(
        self,
        data_mover: Operation | SSAValue,
        bounds: Sequence[Operation | SSAValue],
        strides: Sequence[Operation | SSAValue],
    ):
        super().__init__(operands=[data_mover, bounds, strides])

    @property
    @abstractmethod
    def num(self) -> int:
        raise NotImplementedError()


@irdl_op_definition
class SsrLoop1dOp(SsrLoopBaseOp):
    """
    Configure an SSR data mover for a 1D loop nest.
    """

    name = "snrt.ssr_loop_1d"

    @property
    def num(self) -> int:
        return 1


@irdl_op_definition
class SsrLoop2dOp(SsrLoopBaseOp):
    """
    Configure an SSR data mover for a 2D loop nest.
    """

    name = "snrt.ssr_loop_2d"

    @property
    def num(self) -> int:
        return 2


@irdl_op_definition
class SsrLoop3dOp(SsrLoopBaseOp):
    """
    Configure an SSR data mover for a 3D loop nest.
    """

    name = "snrt.ssr_loop_3d"

    @property
    def num(self) -> int:
        return 3


@irdl_op_definition
class SsrLoop4dOp(SsrLoopBaseOp):
    """
    Configure an SSR data mover for a 4D loop nest.
    """

    name = "snrt.ssr_loop_4d"

    @property
    def num(self) -> int:
        return 4


@irdl_op_definition
class SsrRepeatOp(SnitchRuntimeBaseOperation, ABC):
    """
    Configure the repetition count for a stream.
    """

    name = "snrt.ssr_repeat"
    dm = prop_def(IntegerAttr[IntegerType])
    count = operand_def(i32)

    def __init__(
        self,
        dm: int | IntegerAttr[IntegerType],
        count: Operation | SSAValue,
    ):
        if isinstance(dm, int):
            dm = IntegerAttr(dm, i32)

        super().__init__(operands=[count], properties={"dm": dm})


@irdl_op_definition
class SsrEnableOp(NoOperandNoResultBaseOperation):
    """
    Enable SSR.
    """

    name = "snrt.ssr_enable"


@irdl_op_definition
class SsrDisableOp(NoOperandNoResultBaseOperation):
    """
    Disable SSR.
    """

    name = "snrt.ssr_disable"


class SsrReadWriteBaseOperation(SnitchRuntimeBaseOperation, ABC):
    dm = prop_def(IntegerAttr[IntegerType])
    dim = prop_def(IntegerAttr[IntegerType])
    ptr = operand_def(i32)

    def __init__(
        self,
        dm: int | IntegerAttr[IntegerType],
        dim: int | IntegerAttr[IntegerType],
        ptr: Operation | SSAValue,
    ):
        if isinstance(dm, int):
            dm = IntegerAttr(dm, i32)

        if isinstance(dim, int):
            dim = IntegerAttr(dim, i32)

        super().__init__(
            operands=[ptr],
            properties={
                "dm": dm,
                "dim": dim,
            },
        )


@irdl_op_definition
class SsrReadOp(SsrReadWriteBaseOperation):
    """
    Start a streaming read with a given dimensionality.
    """

    name = "snrt.ssr_read"


@irdl_op_definition
class SsrWriteOp(SsrReadWriteBaseOperation):
    """
    Start a streaming write with a given dimensionality.
    """

    name = "snrt.ssr_write"


@irdl_op_definition
class FpuFenceOp(NoOperandNoResultBaseOperation):
    """
    Synchronize the integer and float pipelines.
    """

    name = "snrt.fpu_fence"


SnitchRuntime = Dialect(
    "snrt",
    [
        GlobalCoreBaseHartidOp,
        GlobalCoreIdxOp,
        GlobalCoreNumOp,
        GlobalComputeCoreIdxOp,
        GlobalComputeCoreNumOp,
        GlobalDmCoreNumOp,
        ClusterCoreIdxOp,
        ClusterCoreNumOp,
        ClusterComputeCoreIdxOp,
        ClusterComputeCoreNumOp,
        ClusterDmCoreIdxOp,
        ClusterDmCoreNumOp,
        ClusterIdxOp,
        ClusterNumOp,
        IsComputeCoreOp,
        IsDmCoreOp,
        ClusterHwBarrierOp,
        ClusterSwBarrierOp,
        GlobalBarrierOp,
        BarrierRegPtrOp,
        GlobalMemoryOp,
        ClusterMemoryOp,
        ZeroMemoryOp,
        DmaStart1DWideptrOp,
        DmaStart1DOp,
        DmaStart2DWideptrOp,
        DmaStart2DOp,
        DmaWaitOp,
        DmaWaitAllOp,
        SsrLoop1dOp,
        SsrLoop2dOp,
        SsrLoop3dOp,
        SsrLoop4dOp,
        SsrRepeatOp,
        SsrEnableOp,
        SsrDisableOp,
        SsrReadOp,
        SsrWriteOp,
        FpuFenceOp,
    ],
    [],
)



from ordered_set import OrderedSet

from xdsl.backend.register_allocator import live_ins_per_block
from xdsl.builder import ImplicitBuilder
from xdsl.dialects.builtin import i32
from xdsl.dialects.test import TestOp
from xdsl.ir import Block, Region, SSAValue


def test_live_ins():
    outer_region = Region(Block(arg_types=(i32, i32)))
    middle_region = Region(Block(arg_types=(i32, i32)))
    inner_region = Region(Block(arg_types=(i32, i32)))
    with ImplicitBuilder(outer_region) as (a, b):
        c, d = TestOp(
            operands=(a, b),
            result_types=(i32, i32),
        ).results
        TestOp(regions=(middle_region,))
        with ImplicitBuilder(middle_region) as (e, f):
            g, h = TestOp(operands=(a, c), result_types=(i32, i32)).results
            TestOp(regions=(inner_region,))
            with ImplicitBuilder(inner_region) as (i, j):
                k, l = TestOp(operands=(b, c, d, g), result_types=(i32, i32)).results
                TestOp(operands=(k,)).results

    name_per_block = {
        outer_region.block: "outer",
        inner_region.block: "inner",
        middle_region.block: "middle",
    }

    name_per_value: dict[SSAValue, str] = {
        a: "a",
        b: "b",
        c: "c",
        d: "d",
        e: "e",
        f: "f",
        g: "g",
        h: "h",
        i: "i",
        j: "j",
        k: "k",
        l: "l",
    }

    # To make test failures easier to read
    def rename(
        mapping: dict[Block, OrderedSet[SSAValue]],
    ) -> dict[str, OrderedSet[str]]:
        return {
            name_per_block[block]: OrderedSet(name_per_value[value] for value in values)
            for block, values in mapping.items()
        }

    reference_live_ins: dict[Block, OrderedSet[SSAValue]] = {
        outer_region.block: OrderedSet([]),
        middle_region.block: OrderedSet([b, c, d, a]),
        inner_region.block: OrderedSet([b, c, d, g]),
    }

    assert rename(live_ins_per_block(inner_region.block)) == rename(
        {inner_region.block: reference_live_ins[inner_region.block]}
    )

    assert rename(live_ins_per_block(middle_region.block)) == rename(
        {
            inner_region.block: reference_live_ins[inner_region.block],
            middle_region.block: reference_live_ins[middle_region.block],
        }
    )

    assert rename(live_ins_per_block(outer_region.block)) == rename(
        {
            inner_region.block: reference_live_ins[inner_region.block],
            middle_region.block: reference_live_ins[middle_region.block],
            outer_region.block: reference_live_ins[outer_region.block],
        }
    )



# RUN: python %s | filecheck %s

from ctypes import c_float, c_int32, c_int64, c_size_t

from xdsl.dialects import builtin
from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.program import FrontendProgram

p = FrontendProgram()
p.register_type(c_int32, builtin.i32)
p.register_type(c_int64, builtin.i64)
p.register_type(c_size_t, builtin.IndexType())
p.register_type(bool, builtin.i1)
p.register_type(c_float, builtin.f32)
p.register_type(float, builtin.f64)
with CodeContext(p):
    # CHECK: @boolean(%{{.*}} : i1)
    def boolean(x: bool):
        return

    # CHECK: @signless(%{{.*}} : i32, %{{.*}} : i64)
    def signless(x: c_int32, y: c_int64):
        return

    # CHECK: @indexed(%{{.*}} : index)
    def indexed(x: c_size_t):
        return

    # CHECK: @fp(%{{.*}} : f32, %{{.*}} : f64)
    def fp(x: c_float, y: float):
        return


p.compile(desymref=False)
print(p.textual_format())



from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import arith, linalg
from xdsl.dialects.builtin import ModuleOp, TensorType
from xdsl.ir import Attribute
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.hints import isa


class LiftAddfPass(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.AddfOp, rewriter: PatternRewriter, /):
        if isa(op.result.type, TensorType[Attribute]):
            rewriter.replace_matched_op(
                linalg.AddOp(op.operands, [op.lhs], [op.result.type])
            )


class LiftSubfPass(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SubfOp, rewriter: PatternRewriter, /):
        if isa(op.result.type, TensorType[Attribute]):
            rewriter.replace_matched_op(
                linalg.SubOp(op.operands, [op.lhs], [op.result.type])
            )


class LiftMulfPass(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.MulfOp, rewriter: PatternRewriter, /):
        if isa(op.result.type, TensorType[Attribute]):
            rewriter.replace_matched_op(
                linalg.MulOp(op.operands, [op.lhs], [op.result.type])
            )


@dataclass(frozen=True)
class LiftArithToLinalg(ModulePass):
    """
    Pass that lifts arith ops to linalg in order to make use of destination-passing style and bufferization.
    """

    name = "lift-arith-to-linalg"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LiftAddfPass(),
                    LiftSubfPass(),
                    LiftMulfPass(),
                ]
            ),
            walk_reverse=False,
            apply_recursively=False,
        )
        module_pass.rewrite_module(op)



import marimo

__generated_with = "0.13.6"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo

    from xdsl.context import Context
    from xdsl.dialects import builtin, arith, func, scf
    from xdsl.utils import marimo as xmo
    from xdsl.printer import Printer
    from typing import Any
    return Any, Context, Printer, arith, builtin, func, mo, scf, xmo


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# MLIR IR""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Textual Format""")
    return


@app.cell(hide_code=True)
def _(mo, triangle_text):
    mo.md(
        rf"""
    MLIR and xDSL use a textual encoding of the IR for debugging, testing, and storing intermediate representations of programs.
    It can be very useful to take a program at some stage of compilation, and inspect it.
    The textual format makes this easy to do.
    Let's look at a representation of a function that sums the first `n` integers:

    {mo.ui.code_editor(triangle_text, language="javascript", disabled=True)}
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""This notebook explains all the components of the above snippet of code. The sections below are structured by _dialect_, which represents a namespace for related abstractions and constructs.""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## The `func` Dialect""")
    return


@app.cell(hide_code=True)
def _():
    swap_text = """\
    func.func @swap(%a : i32, %b : i32) -> (i32, i32) {
      func.return %b, %a : i32, i32
    }"""
    return (swap_text,)


@app.cell(hide_code=True)
def _(mo, swap_text, xmo):
    mo.md(
        rf"""
    The [func dialect](https://mlir.llvm.org/docs/Dialects/Func/) contains building blocks to model function definitions and calls.

    {xmo.module_html(swap_text)}

    The above function takes two 32-bit integers, and returns them in the opposite order.
    In this snippet, there are two operations, `func.func` for function definition and `func.return` to specify the returned values. All operations in MLIR are prefixed with their dialect name.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Exercise 1: Your First Function""")
    return


@app.cell(hide_code=True)
def _():
    first_text = """\
    func.func @first(%arg0: i32, %arg1: i32) -> i32 {
        // Change this to return the second argument instead
        func.return %arg0 : i32
    }\
    """
    return (first_text,)


@app.cell(hide_code=True)
def _(first_text, mo):
    first_text_area = mo.ui.code_editor(first_text, language="javascript")
    return (first_text_area,)


@app.cell(hide_code=True)
def _(exercise_text, first_text_area, mo):
    _first_info_text = exercise_text(
        first_text_area.value,
        "first",
        ((1, 2), (3, 4)),
        ("first(1, 2) = ", "first(3, 4) = "),
    )
    mo.vstack((first_text_area, mo.md(_first_info_text)))
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    second_text = """\
    func.func @first(%arg0: i32, %arg1: i32) -> i32 {
        func.return %arg1 : i32
    }\
    """

    mo.accordion({"Solution": xmo.module_html(second_text)})
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## The `arith` Dialect""")
    return


@app.cell(hide_code=True)
def _():
    add_one_text = """\
    func.func @add_one(%n: i32) -> i32 {
      %one = arith.constant 1 : i32
      %n_plus_one = arith.addi %n, %one : i32
      func.return %n_plus_one : i32
    }"""
    return (add_one_text,)


@app.cell(hide_code=True)
def _(add_one_text, mo, xmo):
    mo.md(
        rf"""
    The [arith dialect](https://mlir.llvm.org/docs/Dialects/ArithOps/) contains arithmetic operations on integers, floating-point values, and other numeric constructs. To start with, here is a function that adds one to its only argument:

    {xmo.module_html(add_one_text)}
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""The `i` in `arith.addi` above stands for integer. Some of the operations, like for addition (`addi`/`addf`), subtraction (`subi`/`subf`), multiplication (`muli`/`mulf`), and others have both integer and floating point variants.""")
    return


@app.cell(hide_code=True)
def _():
    less_than_text = """\
    func.func @less_than(%a: i32, %b: i32) -> i1 {
      %slt = arith.cmpi slt, %lhs, %rhs : i32
      func.return %slt : i1
    }"""
    return (less_than_text,)


@app.cell(hide_code=True)
def _(less_than_text, mo, xmo):
    mo.md(
        rf"""
    The `arith` dialect also contains operations for comparisons. The function below returns the value `true` if a is less than b when the 32-bit values passed in are interpreted as signed integers. Note that the signedness is communicated by the operation itself, not the types of the operands:

    {xmo.module_html(less_than_text)}
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""MLIR IR is in [SSA form](https://en.wikipedia.org/wiki/Static_single-assignment_form), meaning that each value can only be assigned to _once_. This property helps reason about the possible runtime data these values can hold, such as whether they are constant, like the SSA value `%one` in the snippet above.""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Exercise 2: Multiply and Add""")
    return


@app.cell(hide_code=True)
def _():
    fma_text = """\
    func.func @multiply_and_add(%a : i32, %b : i32, %c : i32) -> (i32) {
      // Change this to return a * b + c instead
      func.return %a : i32
    }"""
    return (fma_text,)


@app.cell(hide_code=True)
def _(fma_text, mo):
    fma_text_area = mo.ui.code_editor(fma_text, language="javascript")
    return (fma_text_area,)


@app.cell(hide_code=True)
def _(exercise_text, fma_text_area, mo):
    _fma_info_text = exercise_text(
        fma_text_area.value,
        "multiply_and_add",
        ((1, 2, 3), (4, 5, 6)),
        ("first(1, 2, 3) = ", "first(4, 5, 6) = "),
    )
    mo.vstack((fma_text_area, mo.md(_fma_info_text)))
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    fma_impl = """\
    func.func @multiply_and_add(%a : i32, %b : i32, %c : i32) -> (i32) {
      %ab = arith.muli %a, %b : i32
      %res = arith.addi %ab, %c : i32
      func.return %res : i32
    }"""

    mo.accordion({"Solution": xmo.module_html(fma_impl)})
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## The `scf` Dialect""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""The [`scf` dialect](https://mlir.llvm.org/docs/Dialects/Scf/) contains operations for structured control flow.""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### `scf.if`""")
    return


@app.cell(hide_code=True)
def _():
    select_text = """\
    func.func @select(%cond : i32, %a : i32, %b : i32) -> i32 {
      %res = scf.if %cond -> (i32) {
        scf.yield %a : i32
      } else {
        scf.yield %b : i32
      }
      func.return %res : i32
    }"""
    return (select_text,)


@app.cell(hide_code=True)
def _(mo, select_text, xmo):
    mo.md(
        rf"""
    Here is a function that returns the second argument if the first argument is `true`, and the third argument otherwise:

    {xmo.module_html(select_text)}

    Note that we did not put early returns in the branches of the `scf.if` operation.
    This is due to MLIR's SSA blocks adhering to a specific contract: operations within a block are executed sequentially from top to bottom, and each operation is guaranteed to complete and yield control back to the outer block.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Exercise 3: Abs Function""")
    return


@app.cell(hide_code=True)
def _(abs_function_text, mo):
    abs_input_text = mo.ui.text("-5")
    abs_text_area = mo.ui.code_editor(abs_function_text, language="javascript")
    return abs_input_text, abs_text_area


@app.cell(hide_code=True)
def _():
    abs_function_text = """\
    func.func @abs(%a : i32) -> i32 {
        %false = arith.constant 0 : i1
        %res = scf.if %false -> (i32) {
            scf.yield %a : i32
        } else {
            scf.yield %a : i32
        }
        func.return %res : i32
    }"""
    return (abs_function_text,)


@app.cell(hide_code=True)
def _(Parser, abs_input_text, abs_text_area, ctx, run_func):
    abs_error_text = ""
    abs_results_text = ""
    try:
        abs_input = int(abs_input_text.value)
        abs_inputs = (abs_input,)
        abs_module = Parser(ctx, abs_text_area.value).parse_module()
        abs_results = run_func(abs_module, "abs", abs_inputs)
        abs_results_text = f"abs({abs_input}) = {abs_results}"
    except Exception as e:
        abs_error_text = str(e)

    if abs_error_text:
        abs_info_text = f"""
        Error:

        ```
        {abs_error_text}
        ```
        """
    else:
        abs_info_text = f"""\
        Change the definition of `abs` to compute the absolute value of the input.

        ```
        {abs_results_text}
        ```
        """
    return (abs_info_text,)


@app.cell
def _(abs_info_text, abs_input_text, abs_text_area, mo):
    mo.vstack(
        (
            mo.md(f"""Input: {abs_input_text} {mo.ui.button(label="run")}"""),
            abs_text_area,
            mo.md(abs_info_text),
        )
    )
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    abs_impl = """\
    func.func @abs(%a : i32) -> i32 {
        %zero = arith.constant 0 : i32
        %slt = arith.cmpi slt, %a, %zero : i32
        %res = scf.if %slt -> (i32) {
            %r = arith.subi %zero, %a : i32
            scf.yield %r : i32
        } else {
            scf.yield %a : i32
        }
        func.return %res : i32
    }"""

    mo.accordion({"Solution": xmo.module_html(abs_impl)})
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### `scf.for`""")
    return


@app.cell(hide_code=True)
def _(mo, triangle_text, xmo):
    mo.md(
        rf"""
    The `scf` dialect also contains abstractions to represent for loops, allowing us to implement the triangle function 1 + 2 + 3 + ... + n.

    {xmo.module_html(triangle_text)}

    Due to the SSA contract, we cannot accumulate by updating a value.
    Instead, the loop body takes some number of immutable values and yields the same number of values to use for the next loop.
    When all iterations are complete, these values are returned by the operation.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Exercise 4: Factorial Function""")
    return


@app.cell(hide_code=True)
def _(mo, triangle_text):
    second_input_text = mo.ui.text("5")
    second_text_area = mo.ui.code_editor(
        triangle_text.replace("triangle", "factorial"), language="javascript"
    )
    return second_input_text, second_text_area


@app.cell(hide_code=True)
def _(Parser, ctx, run_func, second_input_text, second_text_area):
    second_error_text = ""
    second_results_text = ""
    try:
        second_input = int(second_input_text.value)
        second_inputs = (second_input,)
        second_module = Parser(ctx, second_text_area.value).parse_module()
        second_results = run_func(second_module, "factorial", second_inputs)
        second_results_text = f"factorial({second_input}) = {second_results}"
    except Exception as e:
        second_error_text = str(e)

    if second_error_text:
        second_info_text = f"""
        Error:

        ```
        {second_error_text}
        ```
        """
    else:
        second_info_text = f"""\
        Change the definition of `factorial` to compute the factorial of the input, instead of the triangle.
        Assume that the input is non-negative.

        ```
        {second_results_text}
        ```
        """
    return (second_info_text,)


@app.cell(hide_code=True)
def _(mo, second_info_text, second_input_text, second_text_area):
    mo.vstack(
        (
            mo.md(f"""Input: {second_input_text} {mo.ui.button(label="run")}"""),
            second_text_area,
            mo.md(second_info_text),
        )
    )
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    fact_impl = """\
    func.func @factorial(%n: i32) -> i32 {
      %zero = arith.constant 0 : i32
      %one = arith.constant 1 : i32
      %n_plus_one = arith.addi %n, %one : i32
      %res = scf.for %i = %one to %n_plus_one step %one iter_args(%acc_in = %one) -> (i32) : i32 {
        %acc_out = arith.muli %acc_in, %i : i32
        scf.yield %acc_out : i32
      }
      func.return %res : i32
    }"""

    mo.accordion({"Solution": xmo.module_html(fact_impl)})
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## The `triangle` revisited""")
    return


@app.cell(hide_code=True)
def _(mo, triangle_text):
    mo.md(
        rf"""
    This notebook contains a very light overview of the most commonly used dialects and operations in MLIR and xDSL, as well as the key concepts of SSA and structured control flow.

    {mo.ui.code_editor(triangle_text, language="javascript", disabled=True)}

    The sections below are a deeper dive into some of the structures that were implicit in the IR snippets we looked at so far, reusing the `triangle` function from earlier.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## The `builtin` Dialect""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    The [`builtin` dialect](https://mlir.llvm.org/docs/Dialects/Builtin/) contains the most commonly-used operation in MLIR and xDSL: the `builtin.module` operation.

    A module is a unit of code which holds a single region.
    The smallest possible piece of code in MLIR IR is an empty module:

    ```
    builtin.module {
    }
    ```

    When the first operation in a file is not a `builtin.module`, it is implicitly assumed and can be omitted, as is the case for all the snippets above.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Attributes""")
    return


@app.cell(hide_code=True)
def _(builtin, mo):
    mo.md(
        rf"""
    Attributes hold compile-time data, such as constants, types, and other information.

    The IR for the `triangle` snippet contains four attributes: `@triangle`, `0`, `1` and `i32`.
    The `i32` attribute is the type of integer values that fit in a register on the target platform.
    As the IR presented here is independent of the machine on which it will be executed on, we do not yet specify the bitwidth of the integer.
    In MLIR, the common way to represent integers of 16, 32, 64, or other bitwidths is, respectively, with the types of `i16`, `i32`, `i64`, etc.
    The `@triangle` attribute is a symbol name, denoting the name of the function.

    There are many other attributes, and a lot of them are part of the `builtin` dialect.
    We will look into defining those in a later notebook, and will only be using attributes from the `builtin` dialect here.

    One important attribute is the dictionary attribute, which looks like this:

    ```
    {builtin.DictionaryAttr({
        "some_string": builtin.StringAttr("my_string"),
        "some_int": builtin.IntegerAttr(42, builtin.i32),
        "some_float": builtin.FloatAttr(3.1415, builtin.f32),
        "a_unit_attr": builtin.UnitAttr()
    })}
    ```

    The last entry denotes a key-value pair (i.e., `a_unit_attr: unit`) where the omitted value is the `unit` attribute.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Generic Format""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    All the snippets presented so far are in IRs that follow the _custom format_, a format that allows operations to specify a pretty and concise representation.
    On the other hand, the _generic format_ is a more uniform and verbose representation that unambiguously shows the structure of an operation.
    Here is the above `triangle` function in generic format:
    """
    )
    return


@app.cell(hide_code=True)
def _(Parser, Printer, StringIO, ctx, mo, triangle_text):
    _triangle_module = Parser(ctx, triangle_text).parse_module()
    _file = StringIO()
    Printer(print_generic_format=True, stream=_file).print(_triangle_module)
    mo.md(
        f"""
    ```
    {_file.getvalue()}
    ```
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""In the next notebooks, we will take a deeper dive into the APIs used to process and construct MLIR IR.""")
    return


@app.cell(hide_code=True)
def _(Printer, StringIO, builtin):
    def print_generic(module: builtin.ModuleOp) -> str:
        io = StringIO()
        Printer(io, print_generic_format=True).print(module)
        return io.getvalue()
    return


@app.cell(hide_code=True)
def _(Context, arith, builtin, func, scf):
    ctx = Context()
    ctx.load_dialect(builtin.Builtin)
    ctx.load_dialect(func.Func)
    ctx.load_dialect(scf.Scf)
    ctx.load_dialect(arith.Arith)
    return (ctx,)


@app.cell(hide_code=True)
def _():
    triangle_text = """\
    func.func @triangle(%n: i32) -> i32 {
      %zero = arith.constant 0 : i32
      %one = arith.constant 1 : i32
      %n_plus_one = arith.addi %n, %one : i32
      %res = scf.for %i = %one to %n_plus_one step %one iter_args(%acc_in = %zero) -> (i32) : i32 {
        %acc_out = arith.addi %acc_in, %i : i32
        scf.yield %acc_out : i32
      }
      func.return %res : i32
    }\
    """
    return (triangle_text,)


@app.cell(hide_code=True)
def _(ctx, triangle_text):
    # parse triangle module to make sure it's valid IR

    from io import StringIO

    from xdsl.parser import Parser, Input

    triangle_module = Parser(ctx, triangle_text, "").parse_module()
    # triangle_module
    return Parser, StringIO


@app.cell(hide_code=True)
def _(Any, builtin):
    def run_func(module: builtin.ModuleOp, name: str, args: tuple[Any, ...]):
        from xdsl.interpreter import Interpreter
        from xdsl.interpreters import scf, arith, func

        interpreter = Interpreter(module)
        interpreter.register_implementations(scf.ScfFunctions)
        interpreter.register_implementations(arith.ArithFunctions)
        interpreter.register_implementations(func.FuncFunctions)

        res = interpreter.call_op(name, args)

        if len(res) == 1:
            res = res[0]

        return res
    return (run_func,)


@app.cell(hide_code=True)
def _(Any, Parser, ctx, run_func):
    def exercise_text(
        module_text: str,
        function_name: str,
        inputs: tuple[Any, ...],
        formats: tuple[str, ...],
    ) -> str:
        error_text = ""
        results_text = ""
        try:
            module = Parser(ctx, module_text).parse_module()
            results_text = "\n".join(
                format + str(run_func(module, function_name, input))
                for format, input in zip(formats, inputs, strict=True)
            )
        except Exception as e:
            error_text = str(e)
        if error_text:
            info_text = f"""
    Error:

    ```
    {error_text}
    ```
    """
        else:
            info_text = f"""
    Here are the outputs when running the function:

    ```
    {results_text}
    ```
    """
        return info_text
    return (exercise_text,)


if __name__ == "__main__":
    app.run()



import argparse
import sys
from collections.abc import Callable, Sequence
from contextlib import redirect_stdout
from importlib.metadata import version
from io import StringIO
from itertools import accumulate
from typing import IO, Any

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass, PassPipeline
from xdsl.printer import Printer
from xdsl.tools.command_line_tool import CommandLineTool
from xdsl.transforms import get_all_passes
from xdsl.utils.exceptions import DiagnosticException, ParseError, ShrinkException
from xdsl.utils.lexer import Span


class xDSLOptMain(CommandLineTool):
    available_passes: dict[str, Callable[[], type[ModulePass]]]
    """
    A mapping from pass names to functions that apply the pass to a ModuleOp.
    """

    available_targets: dict[str, Callable[[ModuleOp, IO[str]], None]]
    """
    A mapping from target names to functions that serialize a ModuleOp into a
    stream.
    """

    pipeline: PassPipeline
    """ The pass-pipeline to be applied. """

    def __init__(
        self,
        description: str = "xDSL modular optimizer driver",
        args: Sequence[str] | None = None,
    ):
        self.available_frontends = {}
        self.available_passes = {}
        self.available_targets = {}

        self.ctx = Context()
        self.register_all_dialects()
        self.register_all_frontends()
        self.register_all_passes()
        self.register_all_targets()

        # arg handling
        arg_parser = argparse.ArgumentParser(description=description)
        self.register_all_arguments(arg_parser)
        self.args = arg_parser.parse_args(args=args)

        self.ctx.allow_unregistered = self.args.allow_unregistered_dialect

        self.setup_pipeline()

    def run(self):
        """
        Executes the different steps.
        """
        chunks, file_extension = self.prepare_input()
        output_stream = self.prepare_output()
        try:
            for i, (chunk, offset) in enumerate(chunks):
                try:
                    if i > 0:
                        output_stream.write("// -----\n")
                    module = self.parse_chunk(chunk, file_extension, offset)

                    if module is not None:
                        if self.apply_passes(module):
                            output_stream.write(self.output_resulting_program(module))
                    output_stream.flush()
                except ParseError as e:
                    s = e.span
                    e.span = Span(s.start, s.end, s.input, offset)
                    if self.args.parsing_diagnostics:
                        print(e)
                    else:
                        raise
                except DiagnosticException as e:
                    if self.args.verify_diagnostics:
                        print(e)
                        # __notes__ only in Python 3.11 and above
                        if hasattr(e, "__notes__"):
                            for e in getattr(e, "__notes__"):
                                print(e)
                    else:
                        raise
                finally:
                    chunk.close()
        except ShrinkException:
            assert self.args.shrink
            print("Success, can shrink")
            # Exit with value 0 to let shrinkray know that it can shrink
            exit(0)
        finally:
            if output_stream is not sys.stdout:
                output_stream.close()
        if self.args.shrink:
            print("Failure, can't shrink")
            # Exit with non-0 value to let shrinkray know that it cannot shrink
            exit(1)

    def register_all_arguments(self, arg_parser: argparse.ArgumentParser):
        """
        Registers all the command line arguments that are used by this tool.

        Add other/additional arguments by overloading this function.
        """
        super().register_all_arguments(arg_parser)

        targets = [name for name in self.available_targets]
        arg_parser.add_argument(
            "-t",
            "--target",
            type=str,
            required=False,
            choices=targets,
            help="target",
            default="mlir",
        )

        arg_parser.add_argument(
            "-o", "--output-file", type=str, required=False, help="path to output file"
        )

        pass_names = ",".join([name for name in self.available_passes])
        arg_parser.add_argument(
            "-p",
            "--passes",
            required=False,
            help=f"Delimited list of passes. Available passes are: {pass_names}",
            type=str,
            default="",
        )

        arg_parser.add_argument(
            "--print-between-passes",
            default=False,
            action="store_true",
            help="Print the IR between each pass",
        )

        arg_parser.add_argument(
            "--verify-diagnostics",
            default=False,
            action="store_true",
            help="Prints the content of a triggered "
            "verifier exception and exits with code 0",
        )

        arg_parser.add_argument(
            "--parsing-diagnostics",
            default=False,
            action="store_true",
            help="Prints the content of a triggered "
            "parsing exception and exits with code 0",
        )

        arg_parser.add_argument(
            "--split-input-file",
            default=False,
            action="store_true",
            help="Split the input file into pieces and process each chunk "
            "independently by using `// -----`",
        )

        arg_parser.add_argument(
            "--print-op-generic",
            default=False,
            action="store_true",
            help="Print operations with the generic format",
        )

        arg_parser.add_argument(
            "--print-no-properties",
            default=False,
            action="store_true",
            help="Print properties as if they were attributes for retrocompatibility.",
        )

        arg_parser.add_argument(
            "--print-debuginfo",
            default=False,
            action="store_true",
            help="Print operations with debug info annotation, such as location.",
        )

        arg_parser.add_argument(
            "-v",
            "--version",
            action=VersionAction,
        )

        arg_parser.add_argument(
            "--shrink",
            default=False,
            action="store_true",
            help="Return success on exit if ShrinkException was raised.",
        )

    def register_pass(
        self, pass_name: str, pass_factory: Callable[[], type[ModulePass]]
    ):
        self.available_passes[pass_name] = pass_factory

    def register_all_passes(self):
        """
        Register all passes that can be used.

        Add other/additional passes by overloading this function.
        """
        for pass_name, pass_factory in get_all_passes().items():
            self.register_pass(pass_name, pass_factory)

    def register_all_targets(self):
        """
        Register all targets that can be used.

        Add other/additional targets by overloading this function.
        """

        def _output_arm_asm(prog: ModuleOp, output: IO[str]):
            from xdsl.dialects.arm import print_assembly

            print_assembly(prog, output)

        def _output_mlir(prog: ModuleOp, output: IO[str]):
            printer = Printer(
                stream=output,
                print_generic_format=self.args.print_op_generic,
                print_properties_as_attributes=self.args.print_no_properties,
                print_debuginfo=self.args.print_debuginfo,
            )
            printer.print_op(prog)
            printer.print_metadata(self.ctx.loaded_dialects)
            print("\n", file=output)

        def _output_riscv_asm(prog: ModuleOp, output: IO[str]):
            from xdsl.dialects.riscv import print_assembly

            print_assembly(prog, output)

        def _output_x86_asm(prog: ModuleOp, output: IO[str]):
            from xdsl.dialects.x86.ops import print_assembly

            print_assembly(prog, output)

        def _output_wat(prog: ModuleOp, output: IO[str]):
            from xdsl.dialects.wasm import WasmModuleOp
            from xdsl.dialects.wasm.wat import WatPrinter

            for op in prog.walk():
                if isinstance(op, WasmModuleOp):
                    printer = WatPrinter(output)
                    op.print_wat(printer)
                    print("", file=output)

        def _emulate_riscv(prog: ModuleOp, output: IO[str]):
            # import only if running riscv emulation
            try:
                from xdsl.interpreters.riscv_emulator import run_riscv
            except ImportError:
                print("Please install optional dependencies to run riscv emulation")
                return

            from xdsl.dialects.riscv import riscv_code

            code = riscv_code(prog)
            with redirect_stdout(output):
                run_riscv(code, unlimited_regs=True, verbosity=0)

        def _output_csl(prog: ModuleOp, output: IO[str]):
            from xdsl.backend.csl.print_csl import print_to_csl

            print_to_csl(prog, output)

        def _output_wgsl(prog: ModuleOp, output: IO[str]):
            from xdsl.backend.wgsl.wgsl_printer import WGSLPrinter
            from xdsl.dialects import gpu

            for op in prog.ops:
                if isinstance(op, gpu.ModuleOp):
                    printer = WGSLPrinter(stream=output)
                    printer.print(op)

        self.available_targets["arm-asm"] = _output_arm_asm
        self.available_targets["csl"] = _output_csl
        self.available_targets["mlir"] = _output_mlir
        self.available_targets["riscemu"] = _emulate_riscv
        self.available_targets["riscv-asm"] = _output_riscv_asm
        self.available_targets["wat"] = _output_wat
        self.available_targets["wgsl"] = _output_wgsl
        self.available_targets["x86-asm"] = _output_x86_asm

    def setup_pipeline(self):
        """
        Creates a pipeline that consists of all the passes specified.

        Fails, if not all passes are registered.
        """

        def callback(
            previous_pass: ModulePass, module: ModuleOp, next_pass: ModulePass
        ) -> None:
            if not self.args.disable_verify:
                module.verify()
            if self.args.print_between_passes:
                print(f"IR after {previous_pass.name}:")
                printer = Printer(stream=sys.stdout)
                printer.print_op(module)
                print("\n\n\n")

        self.pipeline = PassPipeline.parse_spec(
            self.available_passes,
            self.args.passes,
            callback,
        )

    def prepare_input(self) -> tuple[list[tuple[IO[str], int]], str]:
        """
        Prepare input by eventually splitting it in chunks. If not set, the parser
        registered for this file extension is used.
        """

        # when using the split input flag, program is split into multiple chunks
        # it's used for split input file

        chunks: list[tuple[IO[str], int]] = []
        f, file_extension = self.get_input_stream()
        chunks = [(f, 0)]
        if self.args.split_input_file:
            chunks_str = [chunk for chunk in f.read().split("// -----")]
            chunks_off = accumulate(
                [0, *[chunk.count("\n") for chunk in chunks_str[:-1]]]
            )
            chunks = [
                (StringIO(chunk), off)
                for chunk, off in zip(chunks_str, chunks_off, strict=True)
            ]
            f.close()
        if self.args.frontend:
            file_extension = self.args.frontend

        if file_extension not in self.available_frontends:
            for chunk, _ in chunks:
                chunk.close()
            raise ValueError(f"Unrecognized file extension '{file_extension}'")

        return chunks, file_extension

    def prepare_output(self) -> IO[str]:
        if self.args.output_file is None:
            return sys.stdout
        else:
            return open(self.args.output_file, "w")

    def apply_passes(self, prog: ModuleOp) -> bool:
        """Apply passes in order."""
        if not self.args.disable_verify:
            prog.verify()
        self.pipeline.apply(self.ctx, prog)
        if not self.args.disable_verify:
            prog.verify()
        return True

    def output_resulting_program(self, prog: ModuleOp) -> str:
        """Get the resulting program."""
        output = StringIO()
        self.available_targets[self.args.target](prog, output)
        return output.getvalue()


class VersionAction(argparse.Action):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(nargs=0, *args, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        print(f"xdsl-opt built from xdsl version {version('xdsl')}\n")
        parser.exit()



from typing import Any

from xdsl.dialects.printf import PrintFormatOp
from xdsl.interpreter import Interpreter, InterpreterFunctions, impl, register_impls


@register_impls
class PrintfFunctions(InterpreterFunctions):
    @impl(PrintFormatOp)
    def run_println(
        self, interpreter: Interpreter, op: PrintFormatOp, args: tuple[Any, ...]
    ):
        print(op.format_str.data.format(*args), file=interpreter.file)
        return ()



import shutil
import subprocess
from dataclasses import dataclass, field
from io import StringIO

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.parser import Parser
from xdsl.passes import ModulePass
from xdsl.printer import Printer
from xdsl.rewriter import Rewriter
from xdsl.utils.exceptions import DiagnosticException, ParseError


@dataclass(frozen=True)
class MLIROptPass(ModulePass):
    """
    A pass for calling the `mlir-opt` tool with specified parameters. Will fail if
    `mlir-opt` is not available.
    """

    name = "mlir-opt"

    executable: str = field(default="mlir-opt")
    generic: bool = field(default=True)
    arguments: tuple[str, ...] = field(default=())

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        if not shutil.which(self.executable):
            raise ValueError(f"{self.executable} is not available")

        stream = StringIO()
        printer = Printer(print_generic_format=self.generic, stream=stream)
        printer.print_op(op)

        my_string = stream.getvalue()

        completed_process = subprocess.run(
            [self.executable, *self.arguments],
            input=my_string,
            capture_output=True,
            text=True,
        )

        try:
            completed_process.check_returncode()
        except subprocess.CalledProcessError as e:
            raise DiagnosticException("Error executing mlir-opt pass") from e

        # Get the stdout output
        stdout_output = completed_process.stdout
        parser = Parser(ctx, stdout_output)

        try:
            new_module = parser.parse_module()
        except ParseError as e:
            raise DiagnosticException("Error parsing mlir-opt pass output") from e

        op.detach_region(op.body)
        op.add_region(Rewriter().move_region_contents_to_new_regions(new_module.body))
        op.attributes = new_module.attributes



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "xdsl==0.27.0",
# ]
# ///

import marimo

__generated_with = "0.13.6"
app = marimo.App()


@app.cell(hide_code=True)
def _():
    import marimo as mo

    from xdsl.utils import marimo as xmo

    from xdsl.parser import Parser
    from xdsl.context import Context

    from xdsl.dialects.builtin import Builtin
    from xdsl.dialects.arith import Arith
    from xdsl.dialects.func import Func

    from xdsl.ir import Operation

    from xdsl.rewriter import Rewriter
    from xdsl.pattern_rewriter import (
        PatternRewriter,
        RewritePattern,
        GreedyRewritePatternApplier,
        PatternRewriteWalker,
    )
    from xdsl.dialects.arith import AddiOp, ConstantOp, MuliOp
    from xdsl.dialects.builtin import ModuleOp
    from xdsl.dialects.func import FuncOp
    return (
        AddiOp,
        Arith,
        Builtin,
        ConstantOp,
        Context,
        Func,
        GreedyRewritePatternApplier,
        ModuleOp,
        MuliOp,
        Operation,
        Parser,
        PatternRewriteWalker,
        RewritePattern,
        Rewriter,
        mo,
        xmo,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Pattern Rewrites""")
    return


@app.cell(hide_code=True)
def _(Parser, ctx):
    _before_text = """\
    func.func @my_func(%x : index) -> index {
      %c0 = arith.constant 0 : index
      %c2 = arith.constant 2 : index
      %also_x = arith.addi %c0, %x : index
      %two_x = arith.muli %also_x, %c2 : index
      func.return %two_x : index
    }
    """
    before_module = Parser(ctx, _before_text).parse_module()
    None
    return (before_module,)


@app.cell(hide_code=True)
def _(Parser, ctx):
    _after_text = """\
    func.func @my_func(%x : index) -> index {
      %two_x = arith.addi %x, %x : index
      func.return %two_x : index
    }
    """
    after_module = Parser(ctx, _after_text).parse_module()
    None
    return (after_module,)


@app.cell(hide_code=True)
def _(after_module, before_module, mo, xmo):
    mo.md(
        f"""
    Most of the work in compilers is rewriting IR.
    While some transformations have to take modules into account, some can be done locally, such as rewriting an addition with zero to just use the non-zero input.
    These kinds of rewrites are called pattern rewrites, as they apply if some local pattern is matched.

    We can use xDSL's pattern rewriting API to rewrite this function:

    {xmo.module_html(before_module)}

    Into this one:

    {xmo.module_html(after_module)}
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    ## Rationale

    A pattern rewrite is a compiler transformation that matches a DAG in the IR, and replace it with another DAG. For instance, simplifying `x + 0` to `x` is a common optimization that is represented as a pattern rewrite.

    As xDSL and MLIR allow the definition of higher-level dialects, pattern rewrites are very common, and can be used to both write high-level optimizations, and lowerings from a high-level dialect to a lower-level one. A general rationale for pattern rewrites in MLIR can be found [here](https://mlir.llvm.org/docs/Rationale/RationaleGenericDAGRewriter/).

    Pattern rewrites are applied step by step on the IR. For instance, using the rewrite `x + 0 -> x`, the IR `x + 0 + 0` will be progressively transformed to `x + 0`, then `x using this single rewrite. Different application ordering and variations exist for rewrite patterns, which will be covered by Matthias Springer at 4pm on day 3 of the winter school.

    ## Defining a pattern rewrite

    Each Pattern rewrite is a class that inherit from `PatternRewrite`. It defines a single `match_and_rewrite` method, that is called to apply a pattern on an operation. The method will either return without any modification of the IR, or will modify the IR using the `Rewriter`. The most common operation to call on the `Rewriter` is `replace_matched_op`, which will replace the matched operation with a sequence of other operations, and replace the result values of the matched operation with new values.

    Here are two examples of pattern for `x + 0 -> x` and `x * 2 -> x + x`:
    """
    )
    return


@app.cell
def _(
    AddiOp,
    ConstantOp,
    CostantOp,
    MuliOp,
    Operation,
    RewritePattern,
    Rewriter,
):
    # The rewrite x + 0 -> x
    class AddZeroPattern(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: Rewriter):
            # Match an `arith.addi`
            if not isinstance(op, AddiOp):
                return

            # Check if the right hand side is a constant
            if not isinstance(cst := op.rhs.owner, ConstantOp):
                return

            # Check if the constant is 0
            if cst.value.value.data != 0:
                return

            # Replace `x + 0` with `x`
            # The first argument is the new operations to insert
            # The second argument is the replacement for the results of the matched
            # operation (op). As we replace `x + 0` with `x`, we replace the
            # `arith.addi` results with its first argument.
            rewriter.replace_matched_op([], new_results=[op.lhs])

    # The rewrite x * 2 -> x + x
    class MulTwoPattern(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: Rewriter):
            # Match an `arith.addi`
            if not isinstance(op, MuliOp):
                return
            x = op.lhs

            # Check if the right hand side is a constant
            if not isinstance(cst := op.rhs.owner, CostantOp):
                return

            # Check if the constant is 2
            if cst.value.value.data != 2:
                return

            # Replace `x * 2` with `x + x`
            # The results of the `arith.muli` operation are by default the results of the
            # last operation added, so here the `arith.addi`
            add = AddiOp(x, x)
            rewriter.replace_matched_op([add])
    return AddZeroPattern, MulTwoPattern


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    ## Applying rewrite patterns

    There are two steps to apply rewrite patterns in xDSL.

    * First, combining multiple rewrite patterns in a single one, giving priorities between patterns.
    * Then, walking the IR and applying the pattern recursively on all operations until no more pattern can be applied on any operation:

    Here is an example on how to apply rewrites on all operations in a module:
    """
    )
    return


@app.cell
def _(
    AddZeroPattern,
    GreedyRewritePatternApplier,
    ModuleOp,
    MulTwoPattern,
    Operation,
    PatternRewriteWalker,
):
    def apply_all_rewrites(module: ModuleOp, rewrites: list[Operation]):
        merged_pattern = GreedyRewritePatternApplier(AddZeroPattern(), MulTwoPattern())
        walker = PatternRewriteWalker(merged_pattern)
        walker.rewrite_module(module)
    return


@app.cell(hide_code=True)
def _(Arith, Builtin, Context, Func):
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Arith)
    ctx.load_dialect(Func)
    return (ctx,)


if __name__ == "__main__":
    app.run()



from ordered_set import OrderedSet

from xdsl.context import Context
from xdsl.dialects import builtin, eqsat, func
from xdsl.ir import Block, OpResult, Region, SSAValue
from xdsl.passes import ModulePass


class EqsatCreateEgraphsPass(ModulePass):
    """
    Create an egraph from a function by inserting an `eqsat.egraph` operation.

    Input example:
    ```
    func.func @test(%a : index, %b : index) -> (index) {
        %c = arith.addi %a, %b : index
        func.return %c : index
    }
    ```
    Output example:
    ```
    func.func @test(%a : index, %b : index) -> index {
        %c = eqsat.egraph -> index {
            %a_1 = eqsat.eclass %a : index
            %b_1 = eqsat.eclass %b : index
            %c_1 = arith.addi %a_1, %b_1 : index
            %c_2 = eqsat.eclass %c_1 : index
            eqsat.yield %c_2 : index
        }
        func.return %c : index
    }
    ```
    """

    name = "eqsat-create-egraphs"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        for f in op.body.block.ops:
            if isinstance(f, func.FuncOp):
                insert_egraph_op(f)


def insert_egraph_op(f: func.FuncOp):
    egraph_block = Block()
    egraph_body = Region(egraph_block)
    egraph_values: OrderedSet[OpResult] = OrderedSet(())

    def create_eclass(val: SSAValue):
        eclass_op = eqsat.EClassOp(val)
        egraph_block.add_op(eclass_op)
        new_val = eclass_op.results[0]
        egraph_values.add(new_val)
        val.replace_by_if(new_val, lambda u: u.operation is not eclass_op)
        return new_val

    for arg in f.body.block.args:
        create_eclass(arg)

    # we don't walk recursively over all operations, but only
    # the top-level operations in the function body:
    for op in f.body.block.ops:
        if isinstance(op, func.ReturnOp) or not all(
            operand in egraph_values for operand in op.operands
        ):
            continue
        op.detach()
        egraph_block.add_op(op)
        for res in op.results:
            assert res not in egraph_values
            create_eclass(res)

    # Find values that have uses outside the egraph body
    values_to_yield: list[SSAValue] = []
    for val in egraph_values:
        has_external_use = False
        for use in val.uses:
            # Check if the use is outside the egraph block
            if use.operation.parent != egraph_block:
                has_external_use = True
                break
        if has_external_use:
            values_to_yield.append(val)

    # Each value in the egraph block that has a use outside the egraph body should be yielded by the egraph op.
    # Next, these outside uses need to be replaced by the results of the egraph op.
    egraph_block.add_op(eqsat.YieldOp(*values_to_yield))

    # Create the egraph operation with the types of yielded values
    yielded_types = [val.type for val in values_to_yield]
    egraph_op = eqsat.EGraphOp(yielded_types, egraph_body)

    for i, val in enumerate(values_to_yield):
        val.replace_by_if(
            egraph_op.results[i], lambda u: u.operation.parent != egraph_block
        )

    # Insert the egraph operation at the beginning of the function body
    assert f.body.block.first_op is not None, "Function body block is empty"
    f.body.block.insert_op_before(egraph_op, f.body.block.first_op)
